import os
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Any, Callable, List, Optional, Sequence, Type, Union

from tscat import EventQueryInformation, _Catalogue, _Event, add_events_to_catalogue, canonicalize_json_import, \
    canonicalize_votable_import, create_catalogue, create_event, export_json, export_votable, get_catalogues, \
    get_events, import_canonicalized_dict, remove_events_from_catalogue, save
from tscat.filtering import UUID


@dataclass
class Action(ABC):
    user_callback: Optional[Callable[[Any], None]]
    completed: bool = field(init=False)

    def __post_init__(self):
        self.completed = False

    @staticmethod
    def _entity(uuid: str) -> Union[_Catalogue, _Event]:
        entities: Sequence[Union[_Event, _Catalogue]] = get_catalogues(UUID(uuid))
        if len(entities) == 0:
            entities = get_catalogues(UUID(uuid), removed_items=True)
            if len(entities) == 0:
                entities = get_events(UUID(uuid))
                if len(entities) == 0:
                    entities = get_events(UUID(uuid), removed_items=True)

        assert len(entities) == 1

        return entities[0]

    @staticmethod
    def _event(uuid: str) -> _Event:
        entities: Sequence[_Event] = get_events(UUID(uuid))
        if len(entities) == 0:
            entities = get_events(UUID(uuid), removed_items=True)

        assert len(entities) == 1

        return entities[0]

    @abstractmethod
    def action(self):
        ...


@dataclass
class GetCataloguesAction(Action):
    removed_items: bool = False
    catalogues: List[_Catalogue] = field(default_factory=list)

    def action(self) -> None:
        self.catalogues = get_catalogues(removed_items=self.removed_items)


@dataclass
class GetCatalogueAction(Action):
    uuid: str
    removed_items: bool = False
    events: List[_Event] = field(default_factory=list)
    query_info: List[EventQueryInformation] = field(default_factory=list)

    def action(self) -> None:
        catalogues = get_catalogues(UUID(self.uuid))
        if len(catalogues) == 0:
            catalogues = get_catalogues(UUID(self.uuid), removed_items=True)
        self.events, self.query_info = get_events(catalogues[0], removed_items=self.removed_items)


@dataclass
class CreateEntityAction(Action):
    cls: Type[Union[_Catalogue, _Event]]
    args: dict

    entity: Union[_Catalogue, _Event] = field(init=False)

    def action(self) -> None:
        if self.cls == _Catalogue:
            self.entity = create_catalogue(**self.args)
        elif self.cls == _Event:
            self.entity = create_event(**self.args)


@dataclass
class RemoveEntitiesAction(Action):
    uuids: List[str]
    permanently: bool

    def action(self) -> None:
        for entity in map(self._entity, self.uuids):
            entity.remove(permanently=self.permanently)


@dataclass
class AddEventsToCatalogueAction(Action):
    uuids: List[str]
    catalogue_uuid: str

    def action(self) -> None:
        catalogue = self._entity(self.catalogue_uuid)
        assert isinstance(catalogue, _Catalogue)

        # clean uuids of already existing events (assigned, not filtered)
        existing_events_uuids = set(event.uuid for event in get_events(catalogue, assigned_only=True)[0])
        self.uuids = list(set(self.uuids) - existing_events_uuids)
        add_events_to_catalogue(catalogue, list(map(self._event, self.uuids)))


@dataclass
class RemoveEventsFromCatalogueAction(Action):
    uuids: List[str]
    catalogue_uuid: str

    def action(self) -> None:
        catalogue = self._entity(self.catalogue_uuid)
        assert isinstance(catalogue, _Catalogue)
        remove_events_from_catalogue(catalogue, list(map(self._event, self.uuids)))


class EntityChangeAction(Action):
    pass


@dataclass
class SetAttributeAction(EntityChangeAction):
    uuids: List[str]
    name: str
    values: List[Any]

    entities: List = field(default_factory=list)

    def action(self) -> None:
        for uuid, value in zip(self.uuids, self.values):
            entity = self._entity(uuid)
            entity.__setattr__(self.name, value)
            self.entities += [entity]


@dataclass
class DeleteAttributeAction(EntityChangeAction):
    uuids: List[str]
    name: str

    entities: List = field(default_factory=list)

    def action(self) -> None:
        for entity in map(self._entity, self.uuids):
            entity.__delattr__(self.name)
            self.entities += [entity]


@dataclass
class SaveAction(Action):
    def action(self) -> None:
        save()


@dataclass
class MoveToTrashAction(Action):
    uuids: List[str]

    entities: List = field(default_factory=list)

    def action(self) -> None:
        for entity in map(self._entity, self.uuids):
            entity.remove()
            self.entities += [entity]


@dataclass
class RestoreFromTrashAction(Action):
    uuids: List[str]

    entities: List = field(default_factory=list)

    def action(self) -> None:
        for entity in map(self._entity, self.uuids):
            entity.restore()
            self.entities += [entity]


@dataclass
class CanonicalizeImportAction(Action):
    filename: str
    importer: Callable[[str], object] = lambda x: {}

    import_dict: Any = None
    result: Optional[Exception] = None

    def action(self) -> None:
        try:
            self.import_dict = self.importer(self.filename)
        except Exception as e:
            self.result = e


def CanonicalizeJSONImportAction(*args, **kwargs) -> CanonicalizeImportAction:
    return CanonicalizeImportAction(*args, **kwargs,
                                    importer=lambda filename: canonicalize_json_import(open(filename).read()))


def CanonicalizeVOTableImportAction(*args, **kwargs) -> CanonicalizeImportAction:
    from astropy.io.votable import parse
    return CanonicalizeImportAction(*args, **kwargs,
                                    importer=lambda fname: canonicalize_votable_import(
                                        parse(fname),
                                        table_name=os.path.basename(fname).split('.')[0]))


@dataclass
class ImportCanonicalizedDictAction(Action):
    import_dict: Any

    catalogues: List = field(default_factory=list)

    def action(self) -> None:
        self.catalogues = import_canonicalized_dict(self.import_dict)


@dataclass
class ExportAction(Action):
    filename: str
    catalogue_uuids: List[str]
    exporter: Callable[[str, List[_Catalogue]], Optional[Exception]]

    result: Optional[Exception] = None

    def action(self) -> None:
        try:
            catalogues: List[_Catalogue] = list(map(self._entity, self.catalogue_uuids))  # type: ignore
            self.result = self.exporter(self.filename, catalogues)
        except Exception as e:
            self.result = e


def _export_to_json(filename: str, catalogues: List[_Catalogue]) -> Optional[Exception]:
    try:
        with open(filename, 'w+') as f:
            f.write(export_json(catalogues))
            return None
    except Exception as e:
        return e


def ExportJSONAction(*args, **kwargs):
    return ExportAction(*args, **kwargs, exporter=_export_to_json)


def _export_to_votable(filename: str, catalogues: List[_Catalogue]) -> Optional[Exception]:
    try:
        with open(filename, 'w+') as f:
            export_votable(catalogues).to_xml(f)
            return None
    except Exception as e:
        return e


def ExportVotableAction(*args, **kwargs):
    return ExportAction(*args, **kwargs, exporter=_export_to_votable)


@dataclass
class _DeletedEntity:
    type: Union[Type[_Catalogue], Type[_Event]]
    in_trash: bool
    data: dict
    linked_uuids: List[str]
    restored_entity: Optional[Union[_Catalogue, _Event]] = None


@dataclass
class DeletePermanentlyAction(Action):
    uuids: List[str]

    deleted_entities: List[_DeletedEntity] = field(default_factory=list)

    def action(self) -> None:
        for entity in map(self._entity, self.uuids):
            if isinstance(entity, _Catalogue):
                linked_uuids = [e.uuid for e in get_events(entity, assigned_only=True)[0]]
            elif isinstance(entity, _Event):
                linked_uuids = [e.uuid for e in get_catalogues(entity)]

            deleted_entity = _DeletedEntity(
                type(entity),
                entity.is_removed(),
                entity.dump(),
                linked_uuids)
            self.deleted_entities.append(deleted_entity)

            entity.remove(permanently=True)


@dataclass
class RestorePermanentlyDeletedAction(Action):
    deleted_entities: List[_DeletedEntity]

    def action(self) -> None:
        for e in self.deleted_entities:
            entity: Union[_Catalogue, _Event]
            if e.type == _Catalogue:
                entity = create_catalogue(**e.data)
            else:
                entity = create_event(**e.data)

            e.restored_entity = entity

            if e.in_trash:
                entity.remove()

            if e.type == _Catalogue:
                linked_entities = list(map(self._event, e.linked_uuids))
            else:
                linked_entities = list(map(self._entity, e.linked_uuids))

            if isinstance(entity, _Catalogue):
                assert all(lambda x=x: isinstance(x, _Event) for x in linked_entities)
                add_events_to_catalogue(entity, linked_entities)  # type: ignore
            elif isinstance(entity, _Event):
                for c in linked_entities:
                    assert isinstance(c, _Catalogue)
                    add_events_to_catalogue(c, entity)



from typing import List, Optional

from PySide6 import QtCore

from tscat_gui.tscat_driver.actions import ExportJSONAction, ExportVotableAction, ExportAction, \
    CanonicalizeImportAction, \
    CanonicalizeVOTableImportAction, CanonicalizeJSONImportAction
from tscat_gui.undo import Import


def _export(filename: str, uuids: List[str], exporter) -> Optional[Exception]:
    event_loop = QtCore.QEventLoop()

    result = None

    def export_done(action: ExportAction) -> None:
        nonlocal result
        result = action.result

        event_loop.quit()

    from tscat_gui.tscat_driver.model import tscat_model
    tscat_model.do(exporter(export_done, filename, uuids))

    event_loop.exec()

    return result


def export_to_json(filename: str, uuids: List[str]) -> Optional[Exception]:
    return _export(filename, uuids, ExportJSONAction)


def export_to_votable(filename: str, uuids: List[str]) -> Optional[Exception]:
    return _export(filename, uuids, ExportVotableAction)


def _import(filename: str, importer, app_state) -> Optional[Exception]:
    event_loop = QtCore.QEventLoop()

    result = None

    def canonicalization_done(action: CanonicalizeImportAction) -> None:
        nonlocal result
        result = action.result

        if result is None:
            app_state.push_undo_command(Import, filename, action.import_dict)

        event_loop.quit()

    from tscat_gui.tscat_driver.model import tscat_model
    tscat_model.do(importer(canonicalization_done, filename))

    event_loop.exec()

    return result


def import_json(filename: str, app_state) -> Optional[Exception]:
    return _import(filename, CanonicalizeJSONImportAction, app_state)


def import_votable(filename: str, app_state) -> Optional[Exception]:
    return _import(filename, CanonicalizeVOTableImportAction, app_state)



from .flow_layout import FlowLayout
from .editable_label import EditableLabel

from PySide6 import QtWidgets, QtCore, QtGui

import re
import typing

# UTF-8 letters in the beginning, then also numbers and underscore
_tag_validation_regex = re.compile(r'^[a-z][^,]*$')


class _Keyword(EditableLabel):
    closed = QtCore.Signal(QtWidgets.QWidget)

    class DeleteLabel(QtWidgets.QLabel):
        clicked = QtCore.Signal()

        def __init__(self, parent=None):
            super().__init__('✕', parent)
            self.setMargin(1)
            self.setObjectName('DeleteLabel')

        def enterEvent(self, event: QtCore.QEvent) -> None:
            self.setText('✖')
            super().enterEvent(event)  # type: ignore

        def leaveEvent(self, event: QtCore.QEvent) -> None:
            self.setText('✕')
            super().leaveEvent(event)

        def mousePressEvent(self, event: QtGui.QMouseEvent) -> None:
            if event.button() == QtCore.Qt.MouseButton.LeftButton:
                self.clicked.emit()

            super().mousePressEvent(event)

    class Validator(QtGui.QValidator):
        def __init__(self, parent=None):
            super().__init__(parent)

        def validate(self, word: str, pos: int) -> QtGui.QValidator.State:
            if len(word) == 0:
                return QtGui.QValidator.Intermediate  # type: ignore

            if not _tag_validation_regex.match(word):
                return QtGui.QValidator.Intermediate  # type: ignore

            return QtGui.QValidator.Acceptable  # type: ignore

    def __init__(self, text: str, parent: 'EditableKeywordListWidget'):
        super().__init__(text, _Keyword.Validator(), parent)

        self.list_widget = parent

        self.delete_label = _Keyword.DeleteLabel()
        self.delete_label.clicked.connect(lambda: self.closed.emit(self))

        self._layout.addWidget(self.delete_label)
        self._layout.setContentsMargins(5, 5, 5, 5)

        self.setObjectName('OneKeyword')

    def edit(self):
        super().edit()

        completer = QtWidgets.QCompleter(self.list_widget.completion_strings, parent=self)
        completer.setCaseSensitivity(QtCore.Qt.CaseInsensitive)

        self.lineedit.setCompleter(completer)


class EditableKeywordListWidget(QtWidgets.QWidget):
    editingFinished = QtCore.Signal()

    def __init__(self, strings: typing.List[str], completion_strings: typing.List[str] = [], parent=None):
        super().__init__(parent)

        self._layout = FlowLayout()

        self.new_tag = QtWidgets.QToolButton(self)
        self.new_tag.setText('➕')
        self.new_tag.setObjectName('NewTagButton')
        self.new_tag.clicked.connect(lambda _: self._add_tag('', edit=True))  # type: ignore
        self._layout.addWidget(self.new_tag)

        self.tags: typing.List[_Keyword] = []
        for text in strings:
            self._add_tag(text)

        self.completion_strings = set(completion_strings + strings)

        self.setLayout(self._layout)

    def _add_tag(self, text: str, edit=False):
        tag = _Keyword(text, self)
        self.tags += [tag]
        tag.editing_finished.connect(lambda x, _tag=tag: self._tag_text_updated(tag, x))
        tag.closed.connect(self._delete_tag)

        # insert before the new_tag-button
        self._layout.removeWidget(self.new_tag)
        self._layout.addWidget(tag)
        self._layout.addWidget(self.new_tag)

        if edit:
            self._request_edit(tag)

    def _update_tag_texts(self):
        self.editingFinished.emit()

    def _tag_text_updated(self, tag: _Keyword, text: str):
        if len(text) == 0:
            self._delete_tag(tag)
        else:
            self.completion_strings.add(text)
        self._update_tag_texts()

    def _request_edit(self, tag: _Keyword):
        for t in self.tags:
            t.finish_editing()
        tag.edit()

    def _delete_tag(self, tag: _Keyword):
        self.tags.remove(tag)
        self._layout.removeWidget(tag)
        tag.deleteLater()

        self._update_tag_texts()

    def value(self) -> typing.List[str]:
        return [tag.text for tag in self.tags if tag.text]



"""Top-level package for Time series catalogue - GUI."""

__author__ = """Patrick Boettcher"""
__email__ = 'p@yai.se'
__version__ = '0.6.0'

import datetime as dt
import os
from pathlib import Path
from typing import List, Optional, Sequence, Type, Union, cast

import itertools
import sys
from PySide6 import QtCore, QtGui, QtWidgets

from tscat import _Catalogue, _Event
from .edit import EntityEditView
from .model_base.constants import EntityRole, UUIDDataRole
from .state import AppState
from .tscat_driver.actions import Action, AddEventsToCatalogueAction, CanonicalizeImportAction, CreateEntityAction, \
    DeleteAttributeAction, MoveToTrashAction, SaveAction, SetAttributeAction
from .undo import AddEventsToCatalogue, CreateOrSetCataloguePath, DeletePermanently, Import, MoveEntityToTrash, \
    NewCatalogue, NewEvent, RestoreEntityFromTrash, SetAttributeValue
from .utils.import_export import export_to_json, export_to_votable, import_json, import_votable


class _TSCatConcatenateTablePM(QtCore.QConcatenateTablesProxyModel):
    def __init__(self, parent: Optional[QtCore.QObject]):
        super().__init__(parent)

    def indexes_from_uuid(self, uuid: str) -> List[QtCore.QModelIndex]:
        from .tscat_driver.catalog_model import CatalogModel

        indexes = []
        for model in self.sourceModels():
            assert isinstance(model, CatalogModel)
            indexes.append(model.index_from_uuid(uuid))

        indexes = list(map(self.mapFromSource, indexes))
        return indexes


class _TrashAlwaysTopOrBottomSortFilterModel(QtCore.QSortFilterProxyModel):
    def __init__(self, parent=None):
        super().__init__(parent)

    def lessThan(self, source_left: Union[QtCore.QModelIndex, QtCore.QPersistentModelIndex],
                 source_right: Union[QtCore.QModelIndex, QtCore.QPersistentModelIndex]) -> bool:

        left = self.sourceModel().data(source_left)
        right = self.sourceModel().data(source_right)

        if left == 'Trash':
            return False
        elif right == 'Trash':
            return True
        else:
            return left.lower() < right.lower()


class TSCatGUI(QtWidgets.QWidget):
    events_selected = QtCore.Signal(list)
    catalogues_selected = QtCore.Signal(list)
    events_changed = QtCore.Signal(list)
    catalogues_changed = QtCore.Signal(list)

    def __init__(self, parent=None) -> None:
        super().__init__(parent)

        self.state = AppState()

        from .tscat_driver.driver import tscat_driver
        tscat_driver.setParent(self)
        # used a state-variable to differentiate user-induced selections of entities vs programmatically ones
        # needs direct-connected signal to work properly
        self.programmatic_select = False

        self.__setup_ui()
        from .tscat_driver.model import tscat_model
        tscat_model.action_done.connect(self._external_signal_emission_changed)

        self.state.state_changed.connect(self.__state_changed)

    def __state_changed(self, action: str, type: Union[Type[_Catalogue], Type[_Event]],
                        uuids: Sequence[str]) -> None:
        if action in ['active_select', 'passive_select']:
            if type == _Catalogue:
                indexes = list(map(self.catalogue_model.index_from_uuid, uuids))
                indexes = list(map(self.catalogue_sort_filter_model.mapFromSource, indexes))

                if set(self.catalogues_view.selectedIndexes()) != set(indexes):
                    self.programmatic_select = True
                    self.catalogues_view.selectionModel().clear()
                    for index in indexes:
                        self.catalogues_view.selectionModel().select(index,
                                                                     QtCore.QItemSelectionModel.SelectionFlag.Select)
                    self.programmatic_select = False

                current_models = set(self.event_model.sourceModels())
                from .tscat_driver.model import tscat_model
                new_models = set(map(tscat_model.catalog, uuids))

                # temporarily disconnect the selectionChanged signal to avoid unnecessary updates
                self.events_view.selectionModel().selectionChanged.disconnect(  # type: ignore
                    self.__current_event_changed)

                for i in current_models - new_models:
                    self.event_model.removeSourceModel(i)
                for i in new_models - current_models:
                    self.event_model.addSourceModel(i)

                self.events_view.selectionModel().selectionChanged.connect(  # type: ignore
                    self.__current_event_changed, type=QtCore.Qt.DirectConnection)  # type: ignore

            else:
                indexes = list(itertools.chain(*list(map(self.event_model.indexes_from_uuid, uuids))))
                indexes = list(map(self.events_sort_model.mapFromSource, indexes))

                current_selected_indexes_first_column = set(index for index in self.events_view.selectedIndexes()
                                                            if index.column() == 0)
                if current_selected_indexes_first_column != set(indexes):
                    self.programmatic_select = True
                    self.events_view.selectionModel().clear()
                    for index in indexes:
                        self.events_view.selectionModel().select(index,
                                                                 QtCore.QItemSelectionModel.SelectionFlag.Select |
                                                                 QtCore.QItemSelectionModel.SelectionFlag.Rows)
                    self.programmatic_select = False

            if action == 'active_select':
                self._enable_disable_symbol_actions(uuids)

    def _enable_disable_symbol_actions(self, uuids: Sequence[str]) -> None:
        # Enable/disable actions based on the current selection
        self.move_to_trash_action.setEnabled(False)
        self.restore_from_trash_action.setEnabled(False)
        self.delete_action.setEnabled(False)
        self.new_event_action.setEnabled(False)
        self.export_action.setEnabled(False)

        if uuids:
            if len(uuids) == 1:
                self.new_event_action.setEnabled(True)

            enable_restore = False
            enable_move_to_trash = False

            from .tscat_driver.model import tscat_model
            for entity in tscat_model.entities_from_uuids(uuids):
                if entity.is_removed():
                    enable_restore |= True
                else:
                    enable_move_to_trash |= True

            self.restore_from_trash_action.setEnabled(enable_restore)
            self.move_to_trash_action.setEnabled(enable_move_to_trash)
            self.delete_action.setEnabled(True)
            self.export_action.setEnabled(True)

    def __current_event_changed(self, _: QtCore.QModelIndex, __: QtCore.QModelIndex) -> None:
        if not self.programmatic_select:
            uuids = [index.data(UUIDDataRole) for index in self.events_view.selectedIndexes() if index.column() == 0]
            self.state.updated('active_select', _Event, uuids)
            self.events_selected.emit(uuids)

    def __catalogue_selection_changed(self, _: QtCore.QItemSelection, __: QtCore.QItemSelection) -> None:
        if not self.programmatic_select:
            catalogue_uuids = []
            selected_indexes = self.catalogues_view.selectedIndexes()
            for index in selected_indexes:
                if index.data(EntityRole) is not None:  # if not a folder
                    catalogue_uuids.append(index.data(UUIDDataRole))

            if selected_indexes:
                self.state.set_catalogue_path(self.catalogue_model.current_path(selected_indexes[-1]))
            if catalogue_uuids:
                self.state.updated('active_select', _Catalogue, catalogue_uuids)
                self.catalogues_selected.emit(catalogue_uuids)

    def __create_undo_redo_action_menu_on_toolbutton(self,
                                                     index_range: range,
                                                     toolbutton: QtWidgets.QToolButton,
                                                     index_inc: int,
                                                     more_items: bool,
                                                     more_items_text: str) -> None:

        menu = QtWidgets.QMenu(toolbutton)
        for i in index_range:
            action = QtGui.QAction(f'{self.state.undo_stack().command(i).text()}', menu)
            action.triggered[bool].connect(lambda state, _i=i, inc=index_inc:  # type: ignore
                                           self.state.undo_stack().setIndex(_i + inc))
            menu.addAction(action)

        if more_items:
            more = QtGui.QAction(more_items_text, menu)
            more.setEnabled(False)
            menu.addAction(more)

        toolbutton.setMenu(menu)

    def __undo_redo_index_changed(self, index: int) -> None:
        max_action_count = 10

        first_index = max(0, index - max_action_count)
        self.__create_undo_redo_action_menu_on_toolbutton(
            range(index - 1, first_index - 1, -1),
            self.undo_toolbar_button, 0, first_index != 0,
            f'{first_index} more undo actions')

        last_index = min(self.state.undo_stack().count(), index + max_action_count)
        self.__create_undo_redo_action_menu_on_toolbutton(
            range(index, last_index), self.redo_toolbar_button, 1,
            last_index != self.state.undo_stack().count(),
            f'{self.state.undo_stack().count() - last_index} more redo actions')

    def __refresh_current_selection(self) -> None:
        current_selection = self.state.select_state()
        if current_selection.type == _Event:
            self.state.updated('passive_select', _Catalogue, current_selection.selected_catalogues)
        self.state.updated('active_select', current_selection.type, current_selection.selected)

    def __import_from_file(self) -> None:
        filename, filetype = QtWidgets.QFileDialog.getOpenFileName(
            self,
            "Select a catalogue file to be imported",
            str(Path.home()),
            "JSON Document (*.json);;VOTable (*.xml)")

        if filename != '':
            if filetype == 'VOTable (*.xml)':
                result = import_votable(filename, self.state)
            elif filetype == 'JSON Document (*.json)':
                result = import_json(filename, self.state)
            else:
                result = ValueError(f"Unknown filetype '{filetype}'")

            if result:
                QtWidgets.QMessageBox.critical(self,
                                               "Catalogue import",
                                               f"The selected file could not be imported: '{result}'.")

    def __export_to_file(self) -> None:
        filename, filetype = QtWidgets.QFileDialog.getSaveFileName(
            self,
            "Specify the filename for exporting the selected catalogues",
            str(Path.home()),
            "JSON Document (*.json);;VOTable (*.xml)")
        if filename != '':
            if filetype == 'VOTable (*.xml)':
                filename = os.path.splitext(filename)[0] + '.xml'

                result = export_to_votable(filename, self.state.select_state().selected_catalogues)
            elif filetype == 'JSON Document (*.json)':
                filename = os.path.splitext(filename)[0] + '.json'
                result = export_to_json(filename, self.state.select_state().selected_catalogues)
            else:
                result = ValueError(f"Unknown filetype '{filetype}'")

            if result:
                QtWidgets.QMessageBox.critical(
                    self,
                    "Catalogue export",
                    f"The selected catalogues could not be exported to {filename} due to '{result}'.")
            else:
                QtWidgets.QMessageBox.information(self,
                                                  "Catalogue export",
                                                  "The selected catalogues have been successfully exported")

    def __new_folder(self) -> None:
        # use QInputDialog to get the new folder name
        folder_name, ok = QtWidgets.QInputDialog.getText(self, 'Create Folder', 'Folder name:')

        if ok:
            new_path = self.state.current_catalogue_path() + [folder_name]
            node = self.catalogue_model.node_from_path(new_path, True)
            index = self.catalogue_model._index_from_node(node)
            index = self.catalogue_sort_filter_model.mapFromSource(index)

            self.catalogues_view.selectionModel().clear()
            self.catalogues_view.selectionModel().select(index,
                                                         QtCore.QItemSelectionModel.SelectionFlag.Select)
            # expand the new folder
            self.catalogues_view.expand(index)

    def __setup_ui(self) -> None:

        # Event Model and View
        self.event_model = _TSCatConcatenateTablePM(self)

        self.events_sort_model = QtCore.QSortFilterProxyModel()
        self.events_sort_model.setSourceModel(self.event_model)

        self.events_view = QtWidgets.QTableView()
        self.events_view.setMinimumSize(1000, 500)
        self.events_view.setSortingEnabled(True)
        self.events_view.sortByColumn(0, QtCore.Qt.SortOrder.AscendingOrder)

        self.events_view.setModel(self.events_sort_model)
        self.events_view.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectionBehavior.SelectRows)
        self.events_view.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)  # type: ignore
        self.events_view.setDragEnabled(True)
        self.events_view.setDragDropMode(QtWidgets.QTreeView.DragDropMode.DragOnly)

        self.events_view.selectionModel().selectionChanged.connect(  # type: ignore
            self.__current_event_changed, type=QtCore.Qt.DirectConnection)  # type: ignore

        # Edit View
        self.edit_view = EntityEditView(self.state, self)

        # Event/Edit Vertial Splitter
        self.splitter_right = QtWidgets.QSplitter(QtCore.Qt.Vertical, self)  # type: ignore
        self.splitter_right.addWidget(self.events_view)
        self.splitter_right.addWidget(self.edit_view)

        # Catalogue Model and View

        from .tscat_driver.model import tscat_model
        self.catalogue_model = tscat_model.tscat_root()
        self.catalogue_model.events_dropped_on_catalogue.connect(lambda uuid, event_list:
                                                                 self.state.push_undo_command(AddEventsToCatalogue,
                                                                                              uuid,
                                                                                              event_list))
        self.catalogue_model.catalogues_dropped_on_folder.connect(
            lambda catalogues_paths: self.state.push_undo_command(CreateOrSetCataloguePath, catalogues_paths))

        self.catalogues_view = QtWidgets.QTreeView()
        self.catalogues_view.setMinimumSize(300, 900)
        self.catalogues_view.setDragEnabled(True)
        self.catalogues_view.setDragDropMode(QtWidgets.QTreeView.DragDropMode.DragDrop)
        self.catalogues_view.setAcceptDrops(True)
        self.catalogues_view.setDropIndicatorShown(True)
        self.catalogues_view.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)  # type: ignore

        self.catalogue_sort_filter_model = _TrashAlwaysTopOrBottomSortFilterModel()
        self.catalogue_sort_filter_model.setSourceModel(self.catalogue_model)
        self.catalogue_sort_filter_model.setRecursiveFilteringEnabled(True)
        self.catalogue_sort_filter_model.setFilterCaseSensitivity(
            QtCore.Qt.CaseSensitivity.CaseInsensitive)  # type: ignore

        self.catalogues_view.setModel(self.catalogue_sort_filter_model)
        self.catalogues_view.setSortingEnabled(True)
        self.catalogues_view.sortByColumn(0, QtCore.Qt.SortOrder.AscendingOrder)  # type: ignore

        self.catalogues_view.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)  # type: ignore

        self.catalogues_view.selectionModel().selectionChanged.connect(  # type: ignore
            self.__catalogue_selection_changed,
            type=QtCore.Qt.DirectConnection)  # type: ignore

        self.catalogues_view.expanded.connect(
            lambda index: self.catalogue_model.expanded(self.catalogue_sort_filter_model.mapToSource(index)))
        self.catalogues_view.collapsed.connect(
            lambda index: self.catalogue_model.collapsed(self.catalogue_sort_filter_model.mapToSource(index)))

        # Catalogue Layout and Filter
        hlayout = QtWidgets.QHBoxLayout()
        hlayout.setContentsMargins(0, 0, 0, 0)
        hlayout.addWidget(QtWidgets.QLabel('Filter:'))

        catalogue_filter = QtWidgets.QLineEdit()
        catalogue_filter.textChanged.connect(  # type: ignore
            lambda t: self.catalogue_sort_filter_model.setFilterRegularExpression(t))

        hlayout.addWidget(catalogue_filter)

        layout = QtWidgets.QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addLayout(hlayout)
        layout.addWidget(self.catalogues_view)

        left_widget = QtWidgets.QWidget()
        left_widget.setLayout(layout)

        # MainWindow Catalogue/Right Splitter - Horizonal
        splitter = QtWidgets.QSplitter(QtCore.Qt.Horizontal, self)  # type: ignore
        splitter.addWidget(left_widget)
        splitter.addWidget(self.splitter_right)

        layout = QtWidgets.QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)

        # Toolbar
        toolbar = QtWidgets.QToolBar()

        action = QtGui.QAction(self.style().standardIcon(QtWidgets.QStyle.SP_DirIcon),  # type: ignore
                               "Create Folder", self)
        action.triggered.connect(self.__new_folder)  # type: ignore
        toolbar.addAction(action)

        self.new_folder_action = action

        action = QtGui.QAction(self.style().standardIcon(QtWidgets.QStyle.SP_FileDialogNewFolder),  # type: ignore
                               "Create Catalogue", self)
        action.triggered.connect(lambda: self.state.push_undo_command(NewCatalogue))  # type: ignore
        toolbar.addAction(action)

        action = QtGui.QAction(self.style().standardIcon(QtWidgets.QStyle.SP_FileIcon),  # type: ignore
                               "Create Event", self)
        action.triggered.connect(lambda: self.state.push_undo_command(NewEvent))  # type: ignore
        action.setEnabled(False)
        toolbar.addAction(action)

        self.new_event_action = action

        toolbar.addSeparator()
        action = QtGui.QAction(
            self.style().standardIcon(QtWidgets.QStyle.SP_DialogSaveButton), "Save To Disk",  # type: ignore
            self)

        action.triggered.connect(self.save)  # type: ignore
        toolbar.addAction(action)
        action.setEnabled(False)
        self.state.undo_stack_clean_changed.connect(lambda state, a=action: a.setEnabled(not state))

        toolbar.addSeparator()
        undo_action, redo_action = self.state.create_undo_redo_action()

        undo_action.setIcon(self.style().standardIcon(QtWidgets.QStyle.SP_ArrowBack))  # type: ignore
        undo_action.setShortcut(QtCore.Qt.CTRL | QtCore.Qt.Key_Z)  # type: ignore
        toolbar.addAction(undo_action)
        self.undo_toolbar_button = cast(QtWidgets.QToolButton, toolbar.widgetForAction(undo_action))
        self.undo_toolbar_button.setPopupMode(QtWidgets.QToolButton.MenuButtonPopup)  # type: ignore

        redo_action.setIcon(self.style().standardIcon(QtWidgets.QStyle.SP_ArrowForward))  # type: ignore
        redo_action.setShortcut(QtCore.Qt.CTRL | QtCore.Qt.SHIFT | QtCore.Qt.Key_Z)  # type: ignore
        toolbar.addAction(redo_action)
        self.redo_toolbar_button = cast(QtWidgets.QToolButton, toolbar.widgetForAction(redo_action))
        self.redo_toolbar_button.setPopupMode(QtWidgets.QToolButton.MenuButtonPopup)  # type: ignore

        self.state.undo_stack().indexChanged.connect(self.__undo_redo_index_changed)  # type: ignore

        toolbar.addSeparator()

        action = QtGui.QAction(
            self.style().standardIcon(QtWidgets.QStyle.SP_TrashIcon), "Move to Trash", self)  # type: ignore

        action.triggered.connect(lambda: self.state.push_undo_command(MoveEntityToTrash))  # type: ignore
        action.setEnabled(False)
        toolbar.addAction(action)
        self.move_to_trash_action = action

        action = QtGui.QAction(self.style().standardIcon(QtWidgets.QStyle.SP_DialogResetButton),  # type: ignore
                               "Restore from Trash", self)

        action.triggered.connect(lambda: self.state.push_undo_command(RestoreEntityFromTrash))  # type: ignore
        action.setEnabled(False)
        toolbar.addAction(action)
        self.restore_from_trash_action = action

        action = QtGui.QAction(
            self.style().standardIcon(QtWidgets.QStyle.SP_BrowserStop), "Delete permanently",  # type: ignore
            self)

        action.triggered.connect(lambda: self.state.push_undo_command(DeletePermanently))  # type: ignore
        action.setEnabled(False)
        toolbar.addAction(action)
        self.delete_action = action

        toolbar.addSeparator()

        action = QtGui.QAction(
            self.style().standardIcon(QtWidgets.QStyle.SP_DialogRetryButton), "Refresh",  # type: ignore
            self)

        action.triggered.connect(self.__refresh_current_selection)  # type: ignore
        action.setEnabled(False)
        toolbar.addAction(action)

        self.refresh_action = action

        toolbar.addSeparator()

        action = QtGui.QAction(
            self.style().standardIcon(QtWidgets.QStyle.SP_ArrowUp), "Import Catalogue",  # type: ignore
            self)

        action.triggered.connect(self.__import_from_file)  # type: ignore
        toolbar.addAction(action)

        action = QtGui.QAction(self.style().standardIcon(QtWidgets.QStyle.SP_ArrowDown),  # type: ignore
                               "Export Catalogue",
                               self)

        action.triggered.connect(self.__export_to_file)  # type: ignore
        action.setEnabled(False)
        toolbar.addAction(action)

        self.export_action = action

        layout.addWidget(toolbar)
        layout.addWidget(splitter)
        self.setLayout(layout)

    def _external_signal_emission_changed(self, action: Action) -> None:
        if isinstance(action, (SetAttributeAction, DeleteAttributeAction)):
            if isinstance(action.entities[0], _Catalogue):
                self.catalogues_changed.emit([e.uuid for e in action.entities])
            else:
                self.events_changed.emit([e.uuid for e in action.entities])

    @staticmethod
    def update_event_range(uuid: str, start: dt.datetime, stop: dt.datetime) -> None:
        from .tscat_driver.model import tscat_model
        tscat_model.do(SetAttributeAction(None, [uuid], 'start', [start]))
        tscat_model.do(SetAttributeAction(None, [uuid], 'stop', [stop]))

    @staticmethod
    def create_event(start: dt.datetime, stop: dt.datetime, author: str, catalogue_uuid: str) -> _Event:
        event_loop = QtCore.QEventLoop()
        entity = None

        def added_event(_: AddEventsToCatalogueAction) -> None:
            event_loop.quit()

        from .tscat_driver.model import tscat_model

        def creation_callback(action: CreateEntityAction) -> None:
            nonlocal entity
            entity = action.entity
            tscat_model.do(AddEventsToCatalogueAction(added_event, [action.entity.uuid], catalogue_uuid))

        tscat_model.do(CreateEntityAction(creation_callback, _Event,
                                          {
                                              'start': start,
                                              'stop': stop,
                                              'author': author,
                                          }))

        event_loop.exec()

        assert isinstance(entity, _Event)

        return entity

    @staticmethod
    def move_to_trash(uuid: str) -> None:
        from .tscat_driver.model import tscat_model
        tscat_model.do(MoveToTrashAction(None, [uuid]))

    def save(self) -> None:
        def clear_undo_stack(_: SaveAction) -> None:
            self.state.set_undo_stack_clean()

        from .tscat_driver.model import tscat_model
        tscat_model.do(SaveAction(clear_undo_stack))


def main():
    # QtWidgets.QApplication.setDesktopSettingsAware(False)  # defaulting to light mode

    app = QtWidgets.QApplication(sys.argv)
    app.setApplicationName('TSCat GUI')
    app.setApplicationVersion(__version__)
    app.setOrganizationName('Laboratory of Plasma Physics')
    app.setOrganizationDomain('https://www.lpp.fr')
    app.setStyleSheet(
        f"""
        #DeleteLabel {{
                    border-radius: 5px;
                    background: {app.palette().dark().color().name()};
                    min-width: 10px;
                    min-height: 10px;
        }}
        #OneKeyword {{
                border-radius: 15px;
                background: {app.palette().mid().color().name()};
        }}
        """
    )

    main = QtWidgets.QMainWindow()

    w = TSCatGUI(main)

    main.setCentralWidget(w)

    #     styles = """
    # QTreeView::!active { selection-background-color: gray;}
    # """
    #     main.setStyleSheet(styles)

    main.show()

    sys.exit(app.exec())



import typing

from PySide6 import QtCore, QtWidgets, QtGui


class EditableLabel(QtWidgets.QFrame):
    editing_finished = QtCore.Signal(str)
    text_changed = QtCore.Signal(str)

    class Label(QtWidgets.QLabel):
        clicked = QtCore.Signal()

        def __init__(self, text: str, parent=None):
            super().__init__(text, parent)
            self.setObjectName('EditableLabel--Label')

        def mousePressEvent(self, event: QtGui.QMouseEvent) -> None:
            if event.button() == QtCore.Qt.MouseButton.LeftButton:
                self.clicked.emit()
            super().mousePressEvent(event)

        def enterEvent(self, event: QtGui.QEnterEvent) -> None:
            self.setCursor(QtCore.Qt.IBeamCursor)  # type: ignore
            super().enterEvent(event)

        def leaveEvent(self, event: QtCore.QEvent) -> None:
            self.setCursor(QtCore.Qt.ArrowCursor)  # type: ignore
            super().leaveEvent(event)

    class LineEdit(QtWidgets.QLineEdit):
        finished = QtCore.Signal()
        changed = QtCore.Signal(str)
        canceled = QtCore.Signal()

        def __init__(self, text: str, validator: QtGui.QValidator, parent=None):
            super().__init__(text, parent)
            self.setObjectName('EditableLabel--LineEdit')
            self.setValidator(validator)

            self.textChanged.connect(self.text_changed)  # type: ignore
            self.returnPressed.connect(lambda: self.finished.emit())  # type: ignore
            self.editingFinished.connect(lambda: self.finished.emit())  # type: ignore

        def text_changed(self, text: str):
            if self.validator():
                if self.validator().validate(text, len(text) - 1) == QtGui.QValidator.Acceptable:  # type: ignore
                    self.setStyleSheet('border: 1px solid green')
                else:
                    self.setStyleSheet('border: 1px solid red')
            self.changed.emit(text)

        def focusOutEvent(self, event: QtGui.QFocusEvent) -> None:
            # keep focus here if value is not valid
            if self.validator() is None or \
                self.validator().validate(self.text(),
                                          len(self.text()) - 1) == QtGui.QValidator.Acceptable:  # type: ignore
                self.finished.emit()
            else:
                self.canceled.emit()
                # self.setFocus()  maybe it's better to not allow focus-stealing, but seems much harder to implement

            super().focusOutEvent(event)

        def keyPressEvent(self, event: QtGui.QKeyEvent) -> None:
            if event.key() == QtCore.Qt.Key_Escape:  # type: ignore
                self.canceled.emit()

            super().keyPressEvent(event)

    def __init__(self, text: str, validator: typing.Optional[QtGui.QValidator] = None, parent=None):
        super().__init__(parent)
        self.setSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Maximum)  # type: ignore

        self.text = text
        self.validator = validator
        self.label: typing.Optional[EditableLabel.Label] = None
        self.lineedit = None

        self._layout = QtWidgets.QHBoxLayout()
        self._layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(self._layout)

        self.display()

    def display(self):
        if self.label:
            return

        if self.lineedit:
            self._layout.removeWidget(self.lineedit)
            self.lineedit.deleteLater()
            self.lineedit = None

        self.label = EditableLabel.Label(self.text)
        self.label.clicked.connect(self.edit)
        self._layout.insertWidget(0, self.label)

    def edit(self):
        if self.lineedit:
            return

        if self.label:
            self._layout.removeWidget(self.label)
            self.label.deleteLater()
            self.label = None

        self.lineedit = EditableLabel.LineEdit(self.text, self.validator)
        self.lineedit.changed.connect(lambda x: self.text_changed.emit(x))
        self.lineedit.finished.connect(self.finish_editing)
        self.lineedit.canceled.connect(self.cancel_editing)

        self._layout.insertWidget(0, self.lineedit)

        self.lineedit.setFocus()

    def cancel_editing(self):
        self.editing_finished.emit(self.text)
        self.display()

    def finish_editing(self):
        if self.lineedit:
            if self.text != self.lineedit.text():
                self.text = self.lineedit.text()
                self.editing_finished.emit(self.text)
        self.display()



import datetime as dt
import re
from typing import Union, cast, Optional, Sequence

from PySide6 import QtGui, QtWidgets, QtCore

_valid_attribute_name_re = re.compile(r'^[A-Za-z][A-Za-z_0-9]*$')


class AttributeNameValidator(QtGui.QValidator):
    def __init__(self, invalid_words: Sequence[str] = [], parent=None):
        super().__init__(parent)
        self.invalid_words = invalid_words

    def validate(self, word: str, pos: int) -> QtGui.QValidator.State:
        if len(word) == 0:
            return QtGui.QValidator.Intermediate  # type: ignore

        if word in self.invalid_words:
            return QtGui.QValidator.Intermediate  # type: ignore

        if not _valid_attribute_name_re.match(word):
            return QtGui.QValidator.Intermediate  # type: ignore

        return QtGui.QValidator.Acceptable  # type: ignore


class IntDelegate(QtWidgets.QSpinBox):
    def __init__(self, value: Union[int, None] = None, parent: Optional[QtWidgets.QWidget] = None):
        super().__init__(parent)
        self.setRange(-2 ** 31, 2 ** 31 - 1)
        if value is None:
            value = 0
        self.setValue(0 if value is None else value)

    def set_value(self, value: int):
        self.setValue(value)


class FloatDelegate(QtWidgets.QLineEdit):
    def __init__(self, value: Union[float, None] = None, parent: Optional[QtWidgets.QWidget] = None):
        super().__init__(parent)

        validator = QtGui.QDoubleValidator(self)
        validator.setNotation(QtGui.QDoubleValidator.Notation.ScientificNotation)
        self.setValidator(validator)

        if value is None:
            value = 0
        self.setText(str(value))

    def value(self) -> float:
        return float(self.text())

    def set_value(self, value: float):
        self.setText(str(value))


class BoolDelegate(QtWidgets.QCheckBox):
    editingFinished = QtCore.Signal()

    def __init__(self, value: Union[bool, None] = None, parent: Optional[QtWidgets.QWidget] = None):
        super().__init__(parent)
        self.setChecked(False if value is None else value)
        self.stateChanged.connect(lambda: self.editingFinished.emit())  # type: ignore

    def value(self) -> bool:
        return self.isChecked()

    def set_value(self, value: bool):
        self.setChecked(value)


class StrDelegate(QtWidgets.QLineEdit):
    def __init__(self, value: Union[str, None] = None, parent: Optional[QtWidgets.QWidget] = None):
        super().__init__("" if value is None else value, parent)

    def value(self) -> str:
        return self.text()

    def set_value(self, value: str):
        self.setText(value)


class DateTimeDelegate(QtWidgets.QDateTimeEdit):
    def __init__(self, value: Union[dt.datetime, None] = None, parent: Optional[QtWidgets.QWidget] = None):
        super().__init__(dt.datetime.now() if value is None else value, parent)  # type: ignore
        self.setDisplayFormat("yyyy-MM-dd HH:mm:ss:zzz")
        self.setCalendarPopup(True)
        self.setTimeSpec(QtCore.Qt.TimeSpec.UTC)

    def value(self) -> dt.datetime:
        return cast(dt.datetime, self.dateTime().toPython())

    def set_value(self, value: dt.datetime):
        self.setDateTime(value)  # type: ignore



import abc
from abc import ABC
from typing import List, Optional, Sequence

from PySide6 import QtCore

from tscat import _Catalogue, _Event


class Node(ABC):
    def __init__(self) -> None:
        self._parent: Optional['Node'] = None
        self._children: List['Node'] = []

    @property
    def parent(self) -> Optional['Node']:
        return self._parent

    @parent.setter
    def parent(self, parent: 'Node') -> None:
        if self._parent is not None and self._parent != parent:  # reparenting forbidden
            print('reparenting', self, self._parent, parent)
        self._parent = parent

    @property
    def children(self) -> Sequence['Node']:
        return self._children

    def set_children(self, children: Sequence['Node']) -> None:
        self._children = []
        self.append_children(children)

    def append_children(self, children: Sequence['Node']) -> None:
        self._children += children
        for child in children:
            child.parent = self

    def append_child(self, child) -> None:
        self._children.append(child)
        child.parent = self

    def remove_child(self, child) -> None:
        self._children.remove(child)
        child.parent = None

    @property
    def row(self):
        if self._parent is not None:
            return self._parent.children.index(self)
        return 0

    def flags(self):
        return QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsSelectable  # type: ignore

    @property
    @abc.abstractmethod
    def uuid(self) -> str:
        pass


class RootNode(Node):
    def __init__(self):
        super().__init__()

    @property
    def uuid(self) -> str:
        return '00000000-0000-0000-0000-000000000000'

    def flags(self):
        return QtCore.Qt.NoItemFlags


class NamedNode(Node, ABC):
    def __init__(self):
        super().__init__()

    @property
    @abc.abstractmethod
    def name(self) -> str:
        pass


class TrashNode(NamedNode):
    def __init__(self):
        super().__init__()

    @property
    def uuid(self) -> str:
        return '00000000-0000-0000-0000-000000000000'

    @property
    def name(self) -> str:
        return "Trash"

    def flags(self):
        return QtCore.Qt.ItemIsEnabled


class EventNode(Node):
    def __init__(self, event: _Event, assigned: bool) -> None:
        super().__init__()
        self._entity = event
        self._assigned = assigned

    @property
    def uuid(self) -> str:
        return self._entity.uuid

    @property
    def node(self) -> _Event:
        return self._entity

    @node.setter
    def node(self, entity: _Event) -> None:
        self._entity = entity

    def flags(self):
        return super().flags() | QtCore.Qt.ItemIsDragEnabled  # type: ignore

    def is_assigned(self) -> bool:
        return self._assigned


class CatalogNode(NamedNode):
    def __init__(self, catalog: _Catalogue):
        super().__init__()
        self._entity = catalog

    @property
    def uuid(self) -> str:
        return self._entity.uuid

    @property
    def node(self) -> _Catalogue:
        return self._entity

    @node.setter
    def node(self, entity: _Catalogue) -> None:
        self._entity = entity

    @property
    def name(self) -> str:
        return self._entity.name

    def flags(self):
        return super().flags() | QtCore.Qt.ItemIsDragEnabled | QtCore.Qt.ItemIsDropEnabled  # type: ignore


class FolderNode(NamedNode):
    LocalUniqueId: int = 0

    def __init__(self, name: str):
        super().__init__()
        self._name = name
        self._uuid = f'path-uuid-{FolderNode.LocalUniqueId}'
        FolderNode.LocalUniqueId += 1

    def full_path(self) -> List[str]:
        path = []
        node = self
        while isinstance(node, FolderNode):
            path.insert(0, node.name)
            node = node.parent
        return path

    @property
    def uuid(self) -> str:
        return self._uuid

    @property
    def name(self) -> str:
        return self._name

    def flags(self):
        return super().flags() | QtCore.Qt.ItemIsDragEnabled | QtCore.Qt.ItemIsDropEnabled  # type: ignore



import atexit
from typing import Dict, Union, Optional

from PySide6.QtCore import QObject, QThread, Slot, Signal, Qt
from tscat import _Catalogue, _Event

from .actions import Action, GetCataloguesAction, GetCatalogueAction, CreateEntityAction, RemoveEntitiesAction, \
    SetAttributeAction, DeleteAttributeAction, ImportCanonicalizedDictAction


class _TscatDriverWorker(QThread):
    action_done = Signal(Action)

    def __init__(self, parent: Optional[QObject] = None) -> None:
        super().__init__(parent=parent)
        self.moveToThread(self)
        self.start()

    @Slot()
    def do_action(self, action: Action) -> None:
        action.action()
        action.completed = True
        self.action_done.emit(action)


class TscatDriver(QObject):
    # internal signals used between driver and worker
    _do_action = Signal(Action)
    action_done_prioritised = Signal(Action)
    action_done = Signal(Action)

    def __init__(self, parent: Optional[QObject] = None):
        super().__init__(parent=parent)

        self._worker = _TscatDriverWorker()

        self._do_action.connect(self._worker.do_action, Qt.QueuedConnection)
        self._worker.action_done.connect(self._worker_action_done, Qt.QueuedConnection)

        self._entity_cache: Dict[str, Union[_Event, _Catalogue]] = {}
        self.destroyed.connect(self.stop)  # type: ignore

    def do(self, action: Action) -> None:
        self._do_action.emit(action)

    @Slot()
    def _worker_action_done(self, action: Action) -> None:
        if isinstance(action, GetCataloguesAction):
            for catalogue in action.catalogues:
                self._entity_cache[catalogue.uuid] = catalogue

        elif isinstance(action, GetCatalogueAction):
            for event in action.events:
                self._entity_cache[event.uuid] = event

        elif isinstance(action, CreateEntityAction):
            assert action.entity
            self._entity_cache[action.entity.uuid] = action.entity

        elif isinstance(action, RemoveEntitiesAction):
            for uuid in action.uuids:
                if uuid in self._entity_cache:
                    del self._entity_cache[uuid]

        elif isinstance(action, (SetAttributeAction, DeleteAttributeAction)):
            for e in action.entities:
                self._entity_cache[e.uuid] = e

        elif isinstance(action, ImportCanonicalizedDictAction):
            for c in action.catalogues:
                self._entity_cache[c.uuid] = c

        self.action_done_prioritised.emit(action)
        self.action_done.emit(action)

    def entity_from_uuid(self, uuid: str) -> Union[_Event, _Catalogue]:
        return self._entity_cache[uuid]

    def event_from_uuid(self, uuid: str) -> _Event:
        entity = self.entity_from_uuid(uuid)
        assert isinstance(entity, _Event)
        return entity

    def stop(self):
        self._worker.requestInterruption()
        if not self._worker.wait(1000):
            self._worker.quit()


tscat_driver = TscatDriver()
atexit.register(tscat_driver.stop)



from typing import List, cast

import tscat

from .model_base.constants import EntityRole


def __total_event_count(catalogues: List[tscat._Catalogue]) -> str:
    from .tscat_driver.model import tscat_model
    return str(sum(tscat_model.catalog(c.uuid).rowCount() for c in catalogues))


def __global_start_stop_range(catalogues: List[tscat._Catalogue]) -> str:
    from .tscat_driver.model import tscat_model
    min_start, max_stop = None, None
    for c in catalogues:
        model = tscat_model.catalog(c.uuid)
        for i in range(model.rowCount()):
            event = cast(tscat._Event, model.data(model.index(i, 0), EntityRole))

            assert event is not None

            if min_start is None or event.start < min_start:
                min_start = event.start
            if max_stop is None or event.stop > max_stop:
                max_stop = event.stop

    if None in (min_start, max_stop):
        return '-'
    return f'between {min_start} and {max_stop}'


catalogue_meta_data = {
    "Total Events": __total_event_count,
    # "Global Start/Stop": __global_start_stop_range,
}



import datetime as dt
import typing
from typing import Dict, Optional, Type, Union, cast

from PySide6 import QtCore, QtGui, QtWidgets

import tscat.filtering
from tscat.filtering import All, Any, Attribute, Comparison, Field, Has, In, InCatalogue, Match, Not, Predicate
from .utils.editable_label import EditableLabel
from .utils.helper import AttributeNameValidator, BoolDelegate, DateTimeDelegate, FloatDelegate, IntDelegate, \
    StrDelegate


class _PredicateWidget(QtWidgets.QWidget):
    changed = QtCore.Signal()

    def __init__(self, parent: Optional[QtWidgets.QWidget]):
        super().__init__(parent)
        if parent:
            self.changed.connect(lambda p=parent: p.changed.emit())

    def delete_child(self, child):
        pass

    def predicate(self):
        pass


class _Root(_PredicateWidget):
    def __init__(self, predicate: Predicate):
        super().__init__(None)

        self._first = predicate_to_widget_factory(predicate, self)
        if self._first:
            c = QtWidgets.QHBoxLayout()
            c.setContentsMargins(0, 0, 0, 0)
            c.addWidget(self._first)
            self.setLayout(c)

            self._first.changed.connect(lambda: self.changed.emit())

    def delete_child(self, child):
        if self._first:
            self._first.deleteLater()
            self._first = None
            self.changed.emit()

    def predicate(self):
        return self._first.predicate() if self._first else None


class _Comparison(_PredicateWidget):
    OP_LT = 0
    OP_LE = 1
    OP_EQ = 2
    OP_GT = 3
    OP_GE = 4
    OP_NE = 5
    OP_MATCH = 6
    OP_NOT_MATCH = 7

    OP_LABELS = ['<', '<=', '==', '>', '>=', '!=', '=~', '!~']
    OP_LABELS_NEGATED = ['>=', '>', '!=', '<=', '<', '==', '!~', '=~']

    event_fixed_keys_types = {
        'start': dt.datetime,
        'stop': dt.datetime,
        'author': str,
        'uuid': str
    }

    possible_types = {
        'Date': dt.datetime,
        'String': str,
        'Integer': int,
        'Float': float,
        'Bool': bool,
    }

    type_widget = {
        int: IntDelegate,
        str: StrDelegate,
        float: FloatDelegate,
        bool: BoolDelegate,
        dt.datetime: DateTimeDelegate,
    }

    operator_types = {
        '=~': str,
        '!~': str,
    }

    def __init__(self, predicate: Union[Comparison, Match, None], negate: bool, parent: _PredicateWidget):
        super().__init__(parent)

        if predicate is None:
            predicate = Comparison('==', Field('author'), '')
            negate = False

        if type(predicate) == Comparison:
            if negate:
                predicate._op = _Comparison.OP_LABELS_NEGATED[
                    _Comparison.OP_LABELS.index(predicate._op)]  # type: ignore

        self._original_field = predicate._lhs.value
        self._original_value = predicate._rhs

        self._layout = QtWidgets.QHBoxLayout()
        self._layout.setContentsMargins(0, 0, 0, 0)

        self._field = EditableLabel.LineEdit(self._original_field, AttributeNameValidator())
        completer = QtWidgets.QCompleter(list(_Comparison.event_fixed_keys_types.keys()), parent=self)
        completer.setCaseSensitivity(QtCore.Qt.CaseSensitive)  # type: ignore
        self._field.setCompleter(completer)

        self._field.finished.connect(self._field_editing_finished)
        # self._field.canceled.connect(lambda _f=self._field: _f.setText(self._original_field) and self.changed.emit())

        self._layout.addWidget(self._field)

        self._op_label = QtWidgets.QLabel()
        self._op_label.setFont(QtGui.QFont("monospace"))

        if isinstance(predicate, Match):
            op_text = '!~' if negate else '=~'
        else:
            op_text = predicate._op

        self._op_label.setText(op_text)
        self._layout.addWidget(self._op_label)

        self._type_cb = QtWidgets.QComboBox()
        for name, _type in _Comparison.possible_types.items():
            self._type_cb.addItem(name, _type)
        self._type_cb.currentIndexChanged.connect(self._type_index_changed)  # type: ignore
        self._layout.addWidget(self._type_cb)

        self._current_value_widget: Optional[_PredicateWidget] = None
        # used to store widgets which have at some point been used (to not lose last values if type is changed)
        self._value_widgets: Dict[Type, QtWidgets.QWidget] = {}
        self.setLayout(self._layout)

        # initial value of TypeComboBox depends on initial filter-value
        self._select_value_type(type(self._original_value))
        self._type_index_changed(self._type_cb.currentIndex())
        self._current_value_widget.set_value(self._original_value)  # type: ignore

    def _select_value_type(self, selected_type: Union[Type, None] = None):
        value_type_from_field = _Comparison.event_fixed_keys_types.get(self._field.text(), None)
        value_type_from_operator = _Comparison.operator_types.get(self._op_label.text(), None)

        if value_type_from_operator is None and value_type_from_field is None:
            self._type_cb.setEnabled(True)
        else:
            selected_type = value_type_from_operator if value_type_from_operator else value_type_from_field
            self._type_cb.setEnabled(False)

        if selected_type:
            index = self._type_cb.findData(selected_type)
            self._type_cb.setCurrentIndex(index)

    def _field_editing_finished(self):
        self._select_value_type()
        self.changed.emit()

    def _type_index_changed(self, index: int):
        _type = self._type_cb.itemData(index)

        if self._current_value_widget:
            self._layout.removeWidget(self._current_value_widget)
            self._current_value_widget.hide()

        if _type not in self._value_widgets:
            widget = _Comparison.type_widget[_type]()
            widget.editingFinished.connect(lambda: self.changed.emit())  # type: ignore
            self._value_widgets[_type] = widget
        else:
            widget = self._value_widgets[_type]

        self._current_value_widget = widget
        self._layout.addWidget(widget)
        widget.show()
        self.changed.emit()

    def predicate(self) -> Optional[Union[Comparison, Match, Not]]:
        if self._current_value_widget is None:
            return None

        field: Union[Field, Attribute]
        if self._field.text() in tscat._Event._fixed_keys:
            field = Field(self._field.text())
        else:
            field = Attribute(self._field.text())

        if self.operator_index() == _Comparison.OP_MATCH:
            return Match(field, self._current_value_widget.value())  # type: ignore
        elif self.operator_index() == _Comparison.OP_NOT_MATCH:
            return Not(Match(field, self._current_value_widget.value()))  # type: ignore
        else:
            return Comparison(self._op_label.text(), field, self._current_value_widget.value())  # type: ignore

    def set_operator(self, op):
        self._op_label.setText(_Comparison.OP_LABELS[op])
        self._select_value_type()

    def operator_index(self):
        return _Comparison.OP_LABELS.index(self._op_label.text())


class _AttributeIsPresent(_PredicateWidget):
    OP_HAS = 0
    OP_HAS_NOT = 1

    OP_LABELS = ['is present', 'is not present']

    def __init__(self, predicate: Union[Has, None], negate: bool, parent: _PredicateWidget):
        super().__init__(parent)

        c = QtWidgets.QHBoxLayout()
        c.setContentsMargins(0, 0, 0, 0)

        c.addWidget(QtWidgets.QLabel("Attribute"))

        self.attribute = EditableLabel.LineEdit(predicate._operand.value if predicate else "", AttributeNameValidator())
        self.attribute.finished.connect(lambda: self.changed.emit())

        c.addWidget(self.attribute)

        self._op_label = QtWidgets.QLabel()
        self.set_operator(_AttributeIsPresent.OP_HAS_NOT if negate else _AttributeIsPresent.OP_HAS)
        c.addWidget(self._op_label)

        self.setLayout(c)

    def predicate(self) -> Union[Has, Not]:
        predicate = Has(Attribute(self.attribute.text()))
        if self.operator_index() == _AttributeIsPresent.OP_HAS_NOT:
            return Not(predicate)
        return predicate

    def set_operator(self, op):
        self._op_label.setText(_AttributeIsPresent.OP_LABELS[op])

    def operator_index(self):
        return _AttributeIsPresent.OP_LABELS.index(self._op_label.text())


class _StringInStringList(_PredicateWidget):
    OP_IN = 0
    OP_NOT_IN = 1

    OP_LABELS = ['is in string-list', 'is not in string-list']

    event_fixed_keys = ['tags', 'products']

    def __init__(self, predicate: Union[In, None], negate: bool, parent: _PredicateWidget):
        super().__init__(parent)

        c = QtWidgets.QHBoxLayout()
        c.setContentsMargins(0, 0, 0, 0)

        c.addWidget(QtWidgets.QLabel("String"))

        self.le = QtWidgets.QLineEdit(predicate._lhs if predicate else "")
        c.addWidget(self.le)

        self._op_label = QtWidgets.QLabel()
        self.set_operator(_StringInStringList.OP_NOT_IN if negate else _StringInStringList.OP_IN)
        c.addWidget(self._op_label)

        self.attribute = EditableLabel.LineEdit(predicate._rhs.value if predicate else "", AttributeNameValidator())
        completer = QtWidgets.QCompleter(_StringInStringList.event_fixed_keys, parent=self)
        completer.setCaseSensitivity(QtCore.Qt.CaseSensitive)  # type: ignore
        self.attribute.setCompleter(completer)
        self.attribute.finished.connect(lambda: self.changed.emit())

        c.addWidget(self.attribute)

        self.setLayout(c)

    def predicate(self) -> Union[In, Not]:
        attr_or_field_type: Union[Type[Field], Type[
            Attribute]] = Field if self.attribute.text() in _StringInStringList.event_fixed_keys else Attribute
        predicate = In(self.le.text(), attr_or_field_type(self.attribute.text()))
        if self.operator_index() == _StringInStringList.OP_NOT_IN:
            return Not(predicate)
        return predicate

    def set_operator(self, op):
        self._op_label.setText(_StringInStringList.OP_LABELS[op])

    def operator_index(self):
        return _StringInStringList.OP_LABELS.index(self._op_label.text())


class _InCatalogue(_PredicateWidget):
    OP_IN = 0
    OP_NOT_IN = 1

    OP_LABELS = ['is in catalogue', 'is not in catalogue']

    def __init__(self, predicate: Optional[InCatalogue], negate: bool, parent: _PredicateWidget):
        super().__init__(parent)

        c = QtWidgets.QHBoxLayout()
        c.setContentsMargins(0, 0, 0, 0)

        self._op_label = QtWidgets.QLabel()
        self.set_operator(_InCatalogue.OP_NOT_IN if negate else _InCatalogue.OP_IN)
        c.addWidget(self._op_label)

        self.catalogues = QtWidgets.QComboBox()

        from .tscat_driver.model import tscat_model
        from .tscat_driver.nodes import CatalogNode

        for cat in tscat_model.tscat_root().catalogue_nodes(in_trash=False):
            if isinstance(cat, CatalogNode):
                self.catalogues.addItem(cat.name, cat.node)

        for cat in tscat_model.tscat_root().catalogue_nodes(in_trash=True):
            if isinstance(cat, CatalogNode):
                self.catalogues.addItem(cat.name + " (in trash)", cat.node)

        if predicate is not None:
            for index in range(self.catalogues.count()):
                catalogue = cast(tscat._Catalogue, self.catalogues.itemData(index))
                if catalogue and predicate.catalogue and predicate.catalogue.uuid == catalogue.uuid:
                    break
            else:
                index = -1

            if index != -1:
                self.catalogues.setCurrentIndex(index)

        self.catalogues.currentIndexChanged.connect(lambda: self.changed.emit())  # type: ignore

        c.addWidget(self.catalogues)

        self.setLayout(c)

    def predicate(self) -> Union[InCatalogue, Not]:
        predicate = InCatalogue(self.catalogues.itemData(self.catalogues.currentIndex()))
        if self.operator_index() == _InCatalogue.OP_NOT_IN:
            return Not(predicate)
        return predicate

    def set_operator(self, op):
        self._op_label.setText(_InCatalogue.OP_LABELS[op])

    def operator_index(self):
        return _InCatalogue.OP_LABELS.index(self._op_label.text())


operators = [
    ['Equal', _Comparison, _Comparison.OP_EQ],
    ['Not Equal', _Comparison, _Comparison.OP_NE],
    ['Less Than', _Comparison, _Comparison.OP_LT],
    ['Less Or Equal Than', _Comparison, _Comparison.OP_LE],
    ['Greater Than', _Comparison, _Comparison.OP_GT],
    ['Greater Or Equal Than', _Comparison, _Comparison.OP_GE],
    ['Match Regex', _Comparison, _Comparison.OP_MATCH],
    ['Does Not Match Regex', _Comparison, _Comparison.OP_NOT_MATCH],
    ['In String List', _StringInStringList, _StringInStringList.OP_IN],
    ['Not In String List', _StringInStringList, _StringInStringList.OP_NOT_IN],
    ['In Catalogue', _InCatalogue, _InCatalogue.OP_IN],
    ['Not In Catalogue', _InCatalogue, _InCatalogue.OP_NOT_IN],
    ['Attribute Present', _AttributeIsPresent, _AttributeIsPresent.OP_HAS],
    ['Attribute Not Present', _AttributeIsPresent, _AttributeIsPresent.OP_HAS_NOT],
]


class _Condition(_PredicateWidget):
    def __init__(self, predicate: Union[Comparison, Match, In, Has, InCatalogue, None],
                 negate: bool,
                 parent: _PredicateWidget):
        super().__init__(parent)

        self._hlayout = QtWidgets.QHBoxLayout()
        self._hlayout.setContentsMargins(0, 0, 0, 0)

        self._cb = QtWidgets.QComboBox()
        for op in operators:
            self._cb.addItem(str(op[0]), op)
        self._cb.currentIndexChanged.connect(self._condition_operator_changed)  # type: ignore

        self._hlayout.addWidget(self._cb)

        negate = False
        while isinstance(predicate, Not):
            negate = not negate
            predicate = predicate._operand

        self.condition: Union[_Comparison, _AttributeIsPresent, _StringInStringList, _InCatalogue]

        if isinstance(predicate, Comparison) or isinstance(predicate, Match) or predicate is None:
            self.condition = _Comparison(predicate, negate, self)
        elif isinstance(predicate, Has):
            self.condition = _AttributeIsPresent(predicate, negate, self)
        elif isinstance(predicate, In):
            self.condition = _StringInStringList(predicate, negate, self)
        elif isinstance(predicate, InCatalogue):
            self.condition = _InCatalogue(predicate, negate, self)
        else:
            raise ValueError('Unknown predicate type')

        self.condition.changed.connect(lambda: self.changed.emit())

        self.conditions_seen = {type(self.condition): self.condition}

        for index, op in enumerate(operators):
            if isinstance(self.condition, cast(Type, op[1])) and op[2] == self.condition.operator_index():
                self._cb.setCurrentIndex(index)
                break

        self._hlayout.addWidget(self.condition)

        self._hlayout.addStretch()

        self._hlayout.addWidget(_DeletePredicateWidget(self))

        self.setLayout(self._hlayout)

    def _condition_operator_changed(self, index: int):
        data = self._cb.itemData(index)

        if not isinstance(self.condition, data[1]):
            self._hlayout.removeWidget(self.condition)
            self.condition.hide()

            if data[1] in self.conditions_seen:
                self.condition = self.conditions_seen[data[1]]
            else:
                self.condition = data[1](None, False, self)
                self.condition.set_operator(data[2])
                self.condition.changed.connect(lambda: self.changed.emit())
                self.conditions_seen[type(self.condition)] = self.condition

            self._hlayout.insertWidget(self._hlayout.count() - 2, self.condition)
            self.condition.show()
        else:
            self.condition.set_operator(data[2])

        self.changed.emit()

    def predicate(self):
        return self.condition.predicate()

    def delete_child(self, child):
        raise AssertionError('Should never be called.')


class _LogicalCombination(_PredicateWidget):
    def __init__(self,
                 predicate: Union[All, Any, None],
                 negate: bool,
                 parent: _PredicateWidget):
        super().__init__(parent)

        if predicate is None:
            predicate = Any()

        self._children: typing.List[_PredicateWidget] = []

        self._layout = QtWidgets.QVBoxLayout()
        self._layout.setContentsMargins(20, 0, 0, 0)

        button_group = QtWidgets.QButtonGroup(self)

        sub_layout = QtWidgets.QHBoxLayout()
        sub_layout.setContentsMargins(0, 0, 0, 0)

        self._all = QtWidgets.QRadioButton("All Of")
        button_group.addButton(self._all)
        sub_layout.addWidget(self._all)

        self._any = QtWidgets.QRadioButton("Any Of")
        button_group.addButton(self._any)
        sub_layout.addWidget(self._any)

        sub_layout.addStretch()
        sub_layout.addWidget(_DeletePredicateWidget(self))

        button_group.buttonClicked.connect(lambda x: self.changed.emit())  # type: ignore

        if isinstance(predicate, Any):
            self._any.setChecked(True)
        else:
            self._all.setChecked(True)

        self._layout.addLayout(sub_layout)

        for pred in predicate._predicates:
            widget = predicate_to_widget_factory(pred, self)
            self._layout.addWidget(widget)
            self.add_child_predicate(widget)

        add = _AddPredicateWidget(self)
        add.new.connect(lambda x: self.new(x.selected_type()))
        self._layout.addWidget(add)

        self.setLayout(self._layout)

    def new(self, cls: Union[
            Type['_LogicalCombination'],
            Type[_Condition],
            Type[_Comparison],
            Type[_AttributeIsPresent],
            Type[_StringInStringList],
            Type[_InCatalogue]]):
        t = cls(predicate=None, negate=False, parent=self)
        self._layout.insertWidget(self._layout.count() - 1, t)
        self.add_child_predicate(t)
        self.changed.emit()

    def add_child_predicate(self, widget: _PredicateWidget):
        self._children += [widget]

    def child_predicates(self):
        return self._children[:]

    def predicate(self) -> Union[All, Any]:
        child_predicates = []
        for child in self.child_predicates():
            child_predicates += [child.predicate()]

        if self._any.isChecked():
            return Any(*child_predicates)
        else:
            return All(*child_predicates)

    def delete_child(self, widget):
        widget.deleteLater()
        self._layout.removeWidget(widget)
        self._children.remove(widget)
        self.changed.emit()


def predicate_to_widget_factory(predicate: Predicate, parent: _PredicateWidget) -> \
    Union[_LogicalCombination, _Condition]:
    if isinstance(predicate, Any) or isinstance(predicate, All):
        return _LogicalCombination(predicate, False, parent)
    elif isinstance(predicate, Comparison) or isinstance(predicate, Match) or \
        isinstance(predicate, In) or isinstance(predicate, Has) or isinstance(predicate, InCatalogue):
        return _Condition(predicate, False, parent)

    raise ValueError('Unknown predicate type for factory')


class _AddPredicateWidget(QtWidgets.QWidget):
    new = QtCore.Signal(QtWidgets.QWidget)

    def __init__(self, parent: QtWidgets.QWidget):
        super().__init__(parent)

        l = QtWidgets.QHBoxLayout()
        l.setContentsMargins(0, 0, 0, 0)

        self.type_combobox = QtWidgets.QComboBox()
        self.type_combobox.addItem('Condition', _Condition)
        self.type_combobox.addItem('Logical Combination', _LogicalCombination)

        l.addWidget(self.type_combobox)

        button = QtWidgets.QToolButton()
        button.setText('Add')
        button.clicked.connect(lambda: self.new.emit(self))  # type: ignore
        l.addWidget(button)

        l.addStretch()

        self.setLayout(l)

    def selected_type(self):
        return self.type_combobox.itemData(self.type_combobox.currentIndex())


class _DeletePredicateWidget(QtWidgets.QToolButton):
    def __init__(self, widget: _PredicateWidget):
        super().__init__(widget)

        self.setText('✖')

        self.clicked.connect(lambda: widget.parent().delete_child(widget))  # type: ignore


class SimplePredicateEditDialog(QtWidgets.QDialog):
    def __init__(self, predicate: Optional[Predicate], parent=None):
        super().__init__(parent)

        layout = QtWidgets.QVBoxLayout()

        self.predicate = predicate

        if predicate is None:
            predicate = Any()  # default and limit to LogicalCombination when creating a new one

        self._root = _Root(predicate)

        def predicate_changed():
            self.predicate = self._root.predicate()

        self._root.changed.connect(predicate_changed)

        layout.addWidget(self._root)

        ok = QtWidgets.QPushButton("OK")
        ok.clicked.connect(lambda: self.accept())  # type: ignore

        cancel = QtWidgets.QPushButton("Cancel")
        cancel.clicked.connect(lambda: self.reject())  # type: ignore

        last_line_layout = QtWidgets.QHBoxLayout()
        last_line_layout.addStretch()

        last_line_layout.addWidget(ok)
        last_line_layout.addWidget(cancel)

        layout.addStretch()
        layout.addLayout(last_line_layout)

        self.setLayout(layout)



import datetime as dt
import functools
from typing import Any, Callable, Dict, List, Optional, Type, Union, cast

from PySide6 import QtCore, QtWidgets

import tscat
import tscat.filtering
from .metadata import catalogue_meta_data
from .model_base.constants import PathAttributeName
from .predicate import SimplePredicateEditDialog
from .state import AppState
from .tscat_driver.actions import Action, GetCatalogueAction, SetAttributeAction
from .undo import DeleteAttribute, DeleteAttributeAction, NewAttribute, RenameAttribute, SetAttributeValue
from .utils.editable_label import EditableLabel
from .utils.helper import AttributeNameValidator, BoolDelegate, DateTimeDelegate, FloatDelegate, IntDelegate, \
    StrDelegate
from .utils.keyword_list import EditableKeywordListWidget


class _UuidLabelDelegate(QtWidgets.QWidget):
    editingFinished = QtCore.Signal()  # never emitted

    def __init__(self, value: str, parent: Optional[QtWidgets.QWidget] = None):
        super().__init__(parent)

        label = QtWidgets.QLabel(value, self)
        label.setTextInteractionFlags(QtCore.Qt.TextSelectableByMouse)  # type: ignore
        layout = QtWidgets.QHBoxLayout(self)
        layout.addWidget(label)
        layout.addStretch()
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)


class _ReadOnlyString(str):
    pass


class _PredicateDelegate(QtWidgets.QWidget):
    editingFinished = QtCore.Signal()

    def __init__(self, value: Optional[tscat.filtering.Predicate], parent: Optional[QtWidgets.QWidget] = None):
        super().__init__(parent)

        self.predicate = value

        layout = QtWidgets.QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)

        value_str = str(value)
        if len(value_str) > 50:
            value_str = value_str[:50] + '...'
        self.label = QtWidgets.QLabel(value_str)
        layout.addWidget(self.label)

        button = QtWidgets.QPushButton("Edit")
        button.clicked.connect(self.edit_predicate)  # type: ignore
        layout.addWidget(button)

        button = QtWidgets.QPushButton("Clear")
        button.clicked.connect(self.clear_predicate)  # type: ignore
        layout.addWidget(button)
        layout.addStretch()

        self.setLayout(layout)

    def edit_predicate(self):
        dialog = SimplePredicateEditDialog(self.predicate, self)
        result = dialog.exec_()
        if result == QtWidgets.QDialog.Accepted:
            self.predicate = dialog.predicate
            self.editingFinished.emit()

    def clear_predicate(self):
        self.predicate = None
        self.editingFinished.emit()

    def value(self) -> Union[tscat.filtering.Predicate, None]:
        return self.predicate


class _MultipleDifferentValues(list):
    def __init__(self, attribute: str, *args: List) -> None:
        super().__init__(*args)
        self.attribute = attribute


class _RatingDelegate(QtWidgets.QWidget):
    editingFinished = QtCore.Signal()

    def __init__(self, value: Optional[int], parent: Optional[QtWidgets.QWidget] = None):
        super().__init__(parent)

        self.label = QtWidgets.QLabel(self)
        self.slider = QtWidgets.QSlider(self)
        self.slider.setOrientation(QtCore.Qt.Horizontal)  # type: ignore
        self.slider.setRange(0, 10)
        self.slider.setTickInterval(1)
        self.slider.setTickPosition(QtWidgets.QSlider.TicksBelow)  # type: ignore
        self.slider.valueChanged.connect(self._slider_value_changed)  # type: ignore
        self.slider.sliderReleased.connect(lambda: self.editingFinished.emit())  # type: ignore

        if value:
            self.label.setText(str(value))
            self.slider.setValue(value)
        else:
            self.label.setText(" ")
            self.slider.setValue(0)

        layout = QtWidgets.QHBoxLayout(self)
        layout.addWidget(self.slider)
        layout.addWidget(self.label)
        layout.addStretch()
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)

    def _slider_value_changed(self, value: int) -> None:
        if value == 0:
            self.label.setText(" ")
        else:
            self.label.setText(str(value))

    def value(self) -> Any:
        if self.slider.value() == 0:
            return None
        else:
            return self.slider.value()


_delegate_widget_class_factory = {
    'uuid': _UuidLabelDelegate,
    'predicate': _PredicateDelegate,
    'rating': _RatingDelegate,

    _ReadOnlyString: _UuidLabelDelegate,
    int: IntDelegate,
    str: StrDelegate,
    float: FloatDelegate,
    list: EditableKeywordListWidget,
    bool: BoolDelegate,
    dt.datetime: DateTimeDelegate,
}


class _MultipleDifferentValuesDelegate(QtWidgets.QPushButton):
    editingFinished = QtCore.Signal()

    def __init__(self,
                 values: _MultipleDifferentValues,
                 reset_value_selection: Callable[[List[Any]], Any] = lambda x: x[0],
                 parent: Optional[QtWidgets.QWidget] = None):
        super().__init__('<multiple-values-click-to-reset>', parent)
        assert len(values) > 0

        self.reset_value = reset_value_selection(values)

        self.clicked.connect(lambda x: self.editingFinished.emit())  # type: ignore

    def value(self) -> Any:
        return self.reset_value


class _MultipleDifferentValuesDelegateMin(_MultipleDifferentValuesDelegate):
    def __init__(self,
                 values: _MultipleDifferentValues,
                 parent: Optional[QtWidgets.QWidget] = None):
        super().__init__(values, min, parent)


class _MultipleDifferentValuesDelegateMax(_MultipleDifferentValuesDelegate):
    def __init__(self,
                 values: _MultipleDifferentValues,
                 parent: Optional[QtWidgets.QWidget] = None):
        super().__init__(values, max, parent)


_type_name = {
    'Boolean': bool,
    'DateTime': dt.datetime,
    'Float': float,
    'Integer': int,
    'String': str,
    'Word List': list,
}

_type_name_initial_value = {
    'DateTime': lambda: dt.datetime.now()
}


class AttributesGroupBox(QtWidgets.QGroupBox):
    valuesChanged = QtCore.Signal()

    def create_label(self, text: str) -> QtWidgets.QLabel:  # type: ignore
        return QtWidgets.QLabel(text.title())

    def __init__(self, title: str,
                 state: AppState,
                 parent: Optional[QtWidgets.QWidget] = None) -> None:
        super().__init__(title, parent)

        self.attribute_name_labels: Dict[str, QtWidgets.QLabel] = {}
        self.state = state
        self.values: Dict = {}

        self._layout = QtWidgets.QGridLayout()
        self._layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(self._layout)

    def setup_values(self, values: Dict):
        # clear layout, destroy all widgets
        while True:
            item = self._layout.takeAt(0)
            if item:
                item.widget().deleteLater()
            else:
                break

        self.values = values
        self.attribute_name_labels = {}
        for row, attr in enumerate(values.keys()):

            # special case for UUIDs - hide uuid-attribute
            if attr == 'uuid':
                continue

            label = self.create_label(attr)
            self._layout.addWidget(label, row, 0)

            self.attribute_name_labels[attr] = label

            value = values[attr]

            cls: Type[Union[_MultipleDifferentValuesDelegate, _UuidLabelDelegate, _PredicateDelegate,
            IntDelegate, StrDelegate, FloatDelegate,
            EditableKeywordListWidget, BoolDelegate, DateTimeDelegate]]

            if isinstance(value, _MultipleDifferentValues):
                if attr == 'start':
                    cls = _MultipleDifferentValuesDelegateMin
                elif attr == 'stop':
                    cls = _MultipleDifferentValuesDelegateMax
                else:
                    cls = _MultipleDifferentValuesDelegate
            elif attr in _delegate_widget_class_factory:
                cls = _delegate_widget_class_factory[attr]
            else:
                cls = _delegate_widget_class_factory.get(type(value), QtWidgets.QLabel)

            widget = cls(value, parent=self)

            # the editingFinished-signal is not seen by mypy coming from PySide6
            widget.editingFinished.connect(  # type: ignore
                functools.partial(self._edit_finished_on_widget, widget, attr))  # type: ignore
            self._layout.addWidget(widget, row, 1)

    def _edit_finished_on_widget(self, w: QtWidgets.QWidget, a: str) -> None:

        assert isinstance(w, (_MultipleDifferentValuesDelegate, _PredicateDelegate, _RatingDelegate,
                              IntDelegate, StrDelegate, FloatDelegate,
                              EditableKeywordListWidget, BoolDelegate, DateTimeDelegate))

        self._editing_finished(a, w.value())

    def _editing_finished(self, attr, value) -> None:
        if value != self.values[attr]:
            self.state.push_undo_command(SetAttributeValue, attr, value)
            self.valuesChanged.emit()
            self.values[attr] = value


class FixedAttributesGroupBox(AttributesGroupBox):
    def __init__(self,
                 state: AppState,
                 parent: Optional[QtWidgets.QWidget] = None):
        super().__init__("", state, parent)

        self.setStyleSheet("QGroupBox { font-weight: bold; }")

    def setup(self, entities: List[Union[tscat._Catalogue, tscat._Event]]) -> None:
        values: Dict[str, Any] = {}

        for entity in entities:
            for attr in entity.fixed_attributes().keys():
                value = entity.__dict__[attr]
                if attr in values:
                    if isinstance(values[attr], _MultipleDifferentValues):
                        values[attr].append(value)
                    elif values[attr] != value:
                        values[attr] = _MultipleDifferentValues(attr, [values[attr], value])
                else:
                    values[attr] = value

        super().setup_values(values)


class CustomAttributesGroupBox(AttributesGroupBox):

    def create_label(self, text: str) -> EditableLabel:  # type: ignore
        attrs = self.all_attribute_names[:]
        attrs.remove(text)

        name = EditableLabel(text, AttributeNameValidator(attrs))
        name.editing_finished.connect(lambda x, _text=text: self._attribute_name_changed(_text, x))
        name.text_changed.connect(lambda x, _text=text: self._attribute_name_is_changing(_text, x))

        return name

    def __init__(self,
                 state: AppState,
                 parent: Optional[QtWidgets.QWidget] = None) -> None:
        super().__init__("Custom fields", state, parent)

        self.all_attribute_names: List[str] = []

        self.setStyleSheet("QGroupBox { font-weight: bold; }")

    def setup(self, entities: List[Union[tscat._Catalogue, tscat._Event]]) -> None:
        if len(entities) > 1:
            self.hide()
            return

        self.show()

        entity = entities[0]
        self.all_attribute_names = list(entity.variable_attributes().keys()) + \
                                   list(entity.fixed_attributes().keys())

        attributes = sorted(entity.variable_attributes().keys())
        if PathAttributeName in attributes:  # hide PathAttributeName from Editor
            attributes.remove(PathAttributeName)

        values = {}
        for attr in attributes:
            values[attr] = entity.__dict__[attr]

        super().setup_values(values)

        layout = cast(QtWidgets.QGridLayout, self.layout())

        # add a delete-button to each row
        for row, attr in enumerate(attributes):
            but = QtWidgets.QToolButton()
            but.setText('✖')
            but.clicked.connect(lambda a=attr, x=False: self._delete(a))  # type: ignore
            layout.addWidget(but, row, 2)

        # add the new-attribute-button
        new_section_layout = QtWidgets.QHBoxLayout()
        new_section_layout.setContentsMargins(0, 0, 0, 0)

        self.type_combobox = QtWidgets.QComboBox()
        self.type_combobox.addItems(list(_type_name.keys()))
        new_section_layout.addWidget(self.type_combobox)

        button = QtWidgets.QToolButton()
        button.setText('➕')
        button.clicked.connect(self._new)  # type: ignore
        new_section_layout.addWidget(button)

        new_section_layout.addStretch()

        widget = QtWidgets.QWidget()
        widget.setLayout(new_section_layout)

        row = layout.rowCount()
        layout.addWidget(QtWidgets.QLabel('New'), row, 0)
        layout.addWidget(widget, row, 1)

    def _delete(self, attr) -> None:
        self.state.push_undo_command(DeleteAttribute, attr)

    def _new(self) -> None:
        name = 'attribute{}'

        i = 1
        while name.format(i) in self.values.keys():
            i += 1

        name = name.format(i)

        type_name = self.type_combobox.itemText(self.type_combobox.currentIndex())
        default = _type_name_initial_value.get(type_name, _type_name[type_name])()

        self.state.push_undo_command(NewAttribute, name, default)

    def _attribute_name_changed(self, previous: str, text: str) -> None:
        self.state.push_undo_command(RenameAttribute, previous, text)

    def _attribute_name_is_changing(self, previous: str, text: str) -> None:
        # highlight the existing attribute which is using the same text as name
        for label in self.attribute_name_labels.values():
            label.setStyleSheet('background-color: none')

        if text in self.attribute_name_labels and previous != text:
            self.attribute_name_labels[text].setStyleSheet('color: red')


class CatalogueMetaDataGroupBox(AttributesGroupBox):
    def __init__(self,
                 state: AppState,
                 parent: Optional[QtWidgets.QWidget] = None):
        super().__init__("Catalogue(s) information", state, parent)

        self.setStyleSheet("QGroupBox { font-weight: bold; }")

    def setup(self, entities: List[Union[tscat._Catalogue, tscat._Event]]) -> None:

        catalogues = [entity for entity in entities if isinstance(entity, tscat._Catalogue)]
        if len(catalogues) == 0:
            self.hide()
        else:
            values = {}
            self.show()

            for k, value_func in catalogue_meta_data.items():
                value = value_func(catalogues)
                values[k] = _ReadOnlyString(value)

            super().setup_values(values)


class _EntityEditWidget(QtWidgets.QWidget):
    def __init__(self, state: AppState, parent=None) -> None:
        super().__init__(parent)

        layout = QtWidgets.QVBoxLayout()
        layout.setContentsMargins(5, 5, 5, 5)

        self.meta_data = CatalogueMetaDataGroupBox(state)
        layout.addWidget(self.meta_data)

        self.fixed_attributes = FixedAttributesGroupBox(state)
        layout.addWidget(self.fixed_attributes)

        self.attributes: Optional[CustomAttributesGroupBox]
        self.attributes = CustomAttributesGroupBox(state)
        layout.addWidget(self.attributes)

        layout.addStretch()

        self.setLayout(layout)

    def setup(self, uuids: List[str]) -> None:

        from .tscat_driver.model import tscat_model
        entities = tscat_model.entities_from_uuids(uuids)

        if self.meta_data:
            self.meta_data.setup(entities)
        if self.attributes:
            self.attributes.setup(entities)
        self.fixed_attributes.setup(entities)


class EntityEditView(QtWidgets.QScrollArea):

    def __init__(self, state: AppState, parent=None) -> None:
        super().__init__(parent)

        self.edit: Optional[_EntityEditWidget] = None
        self.state = state
        self.current_uuids: List[str] = []

        self.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOn)  # type: ignore
        self.setWidgetResizable(True)

        self.state.state_changed.connect(self.state_changed)

        from .tscat_driver.model import tscat_model
        tscat_model.action_done.connect(self._model_action_done)

    def state_changed(self, action: str,
                      _: Union[Type[tscat._Catalogue], Type[tscat._Event]],
                      uuids: List[str]) -> None:
        if action == 'active_select':
            if self.current_uuids != uuids:
                if self.edit:
                    self.edit.deleteLater()
                    self.edit = None

                self.current_uuids = uuids

                if len(uuids) > 0:
                    self.edit = _EntityEditWidget(self.state)
                    self.setWidget(self.edit)
                    self.edit.setup(self.current_uuids)
        elif action == 'passive_select':
            pass
        else:
            print('unsupported (old) state-changed', action)

    def _model_action_done(self, action: Action) -> None:
        if self.edit:
            if isinstance(action, GetCatalogueAction):
                if action.uuid in self.current_uuids:
                    self.edit.setup(self.current_uuids)
            elif isinstance(action, (SetAttributeAction, DeleteAttributeAction)):
                if any(entity.uuid in self.current_uuids for entity in action.entities):
                    self.edit.setup(self.current_uuids)



import pickle
from typing import Any, List, Optional, Sequence, Union

from PySide6.QtCore import QAbstractTableModel, QMimeData, QModelIndex, QPersistentModelIndex, Qt
from PySide6.QtGui import QColor

from tscat import _Event, _Catalogue
from .actions import Action, AddEventsToCatalogueAction, DeleteAttributeAction, DeletePermanentlyAction, \
    GetCatalogueAction, MoveToTrashAction, RemoveEntitiesAction, RemoveEventsFromCatalogueAction, \
    RestoreFromTrashAction, RestorePermanentlyDeletedAction, SetAttributeAction
from .driver import tscat_driver
from .nodes import CatalogNode, EventNode, TrashNode
from ..model_base.constants import EntityRole, UUIDDataRole
from .mime import MimeType


class CatalogModel(QAbstractTableModel):
    _columns = ['start', 'stop', 'author', 'tags', 'products', 'rating']

    def __init__(self, root: CatalogNode):
        super().__init__()
        self._root = root
        self._trash = TrashNode()
        tscat_driver.action_done_prioritised.connect(self._driver_action_done, Qt.QueuedConnection)

    def _driver_action_done(self, action: Action) -> None:
        if isinstance(action, GetCatalogueAction):
            if action.uuid == self._root.uuid:
                children = [EventNode(e, i.assigned) for e, i in zip(action.events, action.query_info)]
                if action.removed_items:
                    # unused for now
                    self._trash.set_children(children)
                else:
                    self.beginResetModel()
                    self._root.set_children(children)
                    self.endResetModel()

        elif isinstance(action, (SetAttributeAction, DeleteAttributeAction)):
            for e in filter(lambda x: isinstance(x, _Event), action.entities):
                for node in (self._root, self._trash):
                    for row, child in enumerate(node.children):
                        if isinstance(child, EventNode) and child.uuid == e.uuid:
                            child.node = e

                            if node == self._root:  # update model only for visible nodes
                                index_left = self.index(row, 0, QModelIndex())
                                index_right = self.index(row, self.columnCount() - 1, QModelIndex())
                                self.dataChanged.emit(index_left, index_right)  # type: ignore

            if action.name == 'predicate':
                for c in filter(lambda x: isinstance(x, _Catalogue), action.entities):
                    if c.uuid == self._root.uuid:
                        self.refresh()

        elif isinstance(action, AddEventsToCatalogueAction):
            if action.catalogue_uuid == self._root.uuid:
                nodes = list(map(lambda x: EventNode(x, True), map(tscat_driver.event_from_uuid, action.uuids)))

                removed_nodes = list(filter(lambda x: x.node.is_removed(), nodes))
                nodes = list(filter(lambda x: not x.node.is_removed(), nodes))

                self.beginInsertRows(QModelIndex(), self.rowCount(), self.rowCount() + len(nodes) - 1)
                self._root.append_children(nodes)
                self.endInsertRows()

                self._trash.append_children(removed_nodes)

        elif isinstance(action, RestorePermanentlyDeletedAction):
            removed_nodes: List[EventNode] = []  # type: ignore
            nodes: List[EventNode] = []  # type: ignore
            for e in action.deleted_entities:
                if e.type == _Event:
                    event_node = EventNode(e.restored_entity, True)
                    if e.restored_entity.is_removed():
                        removed_nodes.append(event_node)
                    else:
                        nodes.append(event_node)

            self.beginInsertRows(QModelIndex(), self.rowCount(), self.rowCount() + len(nodes) - 1)
            self._root.append_children(nodes)
            self.endInsertRows()

            self._trash.append_children(removed_nodes)

        elif isinstance(action, RemoveEventsFromCatalogueAction):
            if action.catalogue_uuid == self._root.uuid:
                for row, c in reversed(list(enumerate(self._root.children))):
                    for e in action.uuids:
                        if c.uuid == e:
                            self.beginRemoveRows(QModelIndex(), row, row)
                            self._root.remove_child(c)
                            self.endRemoveRows()
                for c in self._trash.children[:]:
                    for e in action.uuids:
                        if c.uuid == e:
                            self._trash.remove_child(c)

        elif isinstance(action, (RemoveEntitiesAction, DeletePermanentlyAction)):
            for row, c in reversed(list(enumerate(self._root.children))):
                if c.uuid in action.uuids:
                    self.beginRemoveRows(QModelIndex(), row, row)
                    self._root.remove_child(c)
                    self.endRemoveRows()

            for c in self._trash.children[:]:
                if c.uuid in action.uuids:
                    self._trash.remove_child(c)

        elif isinstance(action, MoveToTrashAction):
            for row, c in reversed(list(enumerate(self._root.children))):
                if c.uuid in action.uuids:
                    self.beginRemoveRows(QModelIndex(), row, row)
                    self._root.remove_child(c)
                    self.endRemoveRows()
                    self._trash.append_child(c)

        elif isinstance(action, RestoreFromTrashAction):
            nodes = []  # type: ignore
            for c in self._trash.children[:]:
                if c.uuid in action.uuids:
                    self._trash.remove_child(c)
                    nodes.append(c)  # type: ignore

            if nodes:
                self.beginInsertRows(QModelIndex(), self.rowCount(), self.rowCount() + len(nodes) - 1)
                self._root.append_children(nodes)
                self.endInsertRows()

    def refresh(self):
        tscat_driver.do(GetCatalogueAction(None, self._root.uuid))

    def headerData(self, section: int, orientation: Qt.Orientation,
                   role: int = Qt.DisplayRole) -> Any:  # type: ignore
        if orientation == Qt.Orientation.Horizontal:
            if role == Qt.DisplayRole:  # type: ignore
                if section < len(self._columns):
                    return self._columns[section].title()
                else:
                    return "Attributes"
        return None

    def columnCount(self, parent: Optional[Union[QModelIndex, QPersistentModelIndex]] = None) -> int:
        return len(self._columns) + 1

    def rowCount(self, parent: Optional[Union[QModelIndex, QPersistentModelIndex]] = None) -> int:
        return len(self._root.children)

    def flags(self, index: QModelIndex) -> Qt.ItemFlags:  # type: ignore
        if index.isValid():
            node = self._root.children[index.row()]
            return node.flags()

        return Qt.NoItemFlags  # type: ignore

    def data(self, index: Union[QModelIndex, QPersistentModelIndex],
             role: int = Qt.DisplayRole):  # type: ignore
        if not index.isValid():
            return None

        child = self._root.children[index.row()]
        assert isinstance(child, EventNode)

        if role == Qt.DisplayRole:  # type: ignore
            if index.column() < len(self._columns):
                key = self._columns[index.column()]
                return str(child.node.__dict__[key])
            else:
                return str(child.node.variable_attributes())
        elif role == Qt.BackgroundRole:  # type: ignore
            if not child.is_assigned():
                return QColor(Qt.lightGray)  # type: ignore
        elif role == UUIDDataRole:
            return child.uuid
        elif role == EntityRole:
            return child.node

        return None

    def index_from_uuid(self, uuid: str, parent=QModelIndex()) -> QModelIndex:
        for row, event in enumerate(self._root.children):
            if event.uuid == uuid:
                return self.index(row, 0)
        return QModelIndex()

    def mimeTypes(self) -> List[str]:
        return super().mimeTypes() + [MimeType.EVENT_UUID_LIST.value]

    def mimeData(self, indexes: Sequence[QModelIndex]) -> QMimeData:
        mime_data = super().mimeData(indexes)
        mime_data.setData(MimeType.EVENT_UUID_LIST.value,
                          pickle.dumps([index.data(UUIDDataRole) for index in indexes if index.column() == 0]))
        return mime_data



import copy
import dataclasses
from typing import List, Type, Union

from PySide6 import QtCore, QtGui

import tscat
from .logger import log


@dataclasses.dataclass
class SelectState:
    selected: List[str]
    type: Union[Type[tscat._Catalogue], Type[tscat._Event]]
    selected_catalogues: List[str]
    catalogue_path: List[str]


class AppState(QtCore.QObject):
    state_changed = QtCore.Signal(str, type, list)

    undo_stack_clean_changed = QtCore.Signal(bool)

    def __init__(self) -> None:
        super().__init__()
        self._select_state = SelectState([], tscat._Catalogue, [], [])

        self._undo_stack = QtGui.QUndoStack()
        self._undo_stack.cleanChanged.connect(lambda x: self.undo_stack_clean_changed.emit(x))  # type: ignore

    def push_undo_command(self, cls, *args) -> None:
        self._undo_stack.push(cls(self, *args))

    def set_undo_stack_clean(self):
        self._undo_stack.setClean()

    def create_undo_redo_action(self):
        return self._undo_stack.createUndoAction(self), self._undo_stack.createRedoAction(self)

    def select_state(self) -> SelectState:
        return copy.deepcopy(self._select_state)

    def undo_stack(self) -> QtGui.QUndoStack:
        return self._undo_stack

    def updated(self, action: str, ty: Union[Type[tscat._Catalogue], Type[tscat._Event]],
                uuids: List[str]) -> None:
        if action == 'active_select':
            if uuids != self._select_state.selected:
                self._select_state.selected = uuids[:]
                self._select_state.type = ty

                if self._select_state.type == tscat._Catalogue:
                    self._select_state.selected_catalogues = uuids[:]
            else:
                log.debug(f'already active "{uuids}"')

        log.debug(f'app-state-updated action:{action}, type:{ty}, uuids:{uuids}')
        self.state_changed.emit(action, ty, uuids)

    def set_catalogue_path(self, path: List[str]) -> None:
        self._select_state.catalogue_path = path

    def current_catalogue_path(self) -> List[str]:
        return self._select_state.catalogue_path



import pickle
from typing import  List

from PySide6.QtCore import  QMimeData

from enum import Enum

class MimeType(Enum):
    CATALOGUE_UUID_LIST = 'application/x-tscat-catalogue-uuid-list'
    EVENT_UUID_LIST = 'application/x-tscat-event-uuid-list'

def is_catalogue_uuid_list(data: QMimeData) -> bool:
    """Check if the QMimeData contains a catalogue UUID list."""
    return data.hasFormat(MimeType.CATALOGUE_UUID_LIST.value)

def is_event_uuid_list(data: QMimeData) -> bool:
    """Check if the QMimeData contains an event UUID list."""
    return data.hasFormat(MimeType.EVENT_UUID_LIST.value)

def catalogue_uuid_list(data: QMimeData) -> List[str]:
    """Extract a list of catalogue UUIDs from the QMimeData."""
    return pickle.loads(data.data(MimeType.CATALOGUE_UUID_LIST.value))  # type: ignore

def event_uuid_list(data: QMimeData) -> List[str]:
    """Extract a list of event UUIDs from the QMimeData."""
    return pickle.loads(data.data(MimeType.EVENT_UUID_LIST.value))  # type: ignore



from typing import Union, Sequence, List

from PySide6.QtCore import QObject, QAbstractItemModel, Signal, Qt
from tscat import _Event, _Catalogue

from .actions import Action
from .tscat_root_model import TscatRootModel


class _Model(QObject):
    action_done = Signal(Action)

    def __init__(self):
        super().__init__(None)

        self._tscat_root = TscatRootModel()

        from .driver import tscat_driver
        tscat_driver.action_done.connect(self._action_done, Qt.QueuedConnection)

    def _action_done(self, action: Action) -> None:
        self.action_done.emit(action)
        if action.user_callback:
            action.user_callback(action)

    def catalog(self, uid: str) -> QAbstractItemModel:
        return self._tscat_root.catalog(uid)

    def tscat_root(self) -> TscatRootModel:
        return self._tscat_root

    @staticmethod
    def do(action: Action) -> None:
        from .driver import tscat_driver
        return tscat_driver.do(action)

    @staticmethod
    def entities_from_uuids(uuids: Sequence[str]) -> List[Union[_Catalogue, _Event]]:
        from .driver import tscat_driver
        return list(map(tscat_driver.entity_from_uuid, uuids))


tscat_model = _Model()



import abc
import datetime as dt
import os
from copy import deepcopy
from typing import Dict, List, Optional, Tuple, Type, Union

from PySide6 import QtGui

import tscat
from .model_base.constants import PathAttributeName
from .state import AppState
from .tscat_driver.actions import AddEventsToCatalogueAction, CreateEntityAction, DeleteAttributeAction, \
    DeletePermanentlyAction, ImportCanonicalizedDictAction, MoveToTrashAction, RemoveEntitiesAction, \
    RemoveEventsFromCatalogueAction, RestoreFromTrashAction, RestorePermanentlyDeletedAction, SetAttributeAction, \
    _DeletedEntity


class _EntityBased(QtGui.QUndoCommand):
    def __init__(self, state: AppState, parent=None) -> None:
        super().__init__(parent)

        self.state = state
        self._select_state = state.select_state()

    def _selected_entities(self) -> List[str]:
        return self._select_state.selected

    def _mapped_selected_entities(self) -> List[Union[tscat._Catalogue, tscat._Event]]:
        from .tscat_driver.model import tscat_model
        return tscat_model.entities_from_uuids(self._selected_entities())

    def _select(self, uuids: List[str],
                type: Optional[Union[Type[tscat._Catalogue], Type[tscat._Event]]] = None):
        if type is None:
            type = self._select_state.type

        if type == tscat._Event:
            self.state.updated('passive_select', tscat._Catalogue, self._select_state.selected_catalogues)
        self.state.updated('active_select', type, uuids)
        self.state.set_catalogue_path(self._select_state.catalogue_path)

    def redo(self) -> None:
        self._redo()

    def undo(self) -> None:
        self._undo()

    @abc.abstractmethod
    def _redo(self) -> None:
        pass

    @abc.abstractmethod
    def _undo(self) -> None:
        pass


class NewAttribute(_EntityBased):
    def __init__(self, state: AppState, name: str, value, parent=None):
        super().__init__(state, parent)

        self.setText(f'Create attribute {name} in {self._select_state.type}')

        self.name = name
        self.value = value

    def _redo(self) -> None:
        from .tscat_driver.model import tscat_model
        tscat_model.do(SetAttributeAction(None, self._selected_entities(),
                                          self.name, [self.value] * len(self._selected_entities())))

        self._select(self._selected_entities())

    def _undo(self) -> None:
        from .tscat_driver.model import tscat_model
        tscat_model.do(DeleteAttributeAction(None, self._selected_entities(), self.name))

        self._select(self._selected_entities())


class RenameAttribute(_EntityBased):
    def __init__(self, state: AppState, old_name: str, new_name: str,
                 parent=None):
        super().__init__(state, parent)

        self.setText(f'Rename attribute from {old_name} to {new_name} in {self._select_state.type}')

        self.new_name = new_name
        self.old_name = old_name

    def _redo(self) -> None:
        values = [entity.__dict__[self.old_name] for entity in self._mapped_selected_entities()]
        from .tscat_driver.model import tscat_model
        tscat_model.do(DeleteAttributeAction(None, self._selected_entities(), self.old_name))
        tscat_model.do(SetAttributeAction(None, self._selected_entities(), self.new_name, values))

        self._select(self._selected_entities())

    def _undo(self) -> None:
        values = [entity.__dict__[self.new_name] for entity in self._mapped_selected_entities()]
        from .tscat_driver.model import tscat_model
        tscat_model.do(DeleteAttributeAction(None, self._selected_entities(), self.new_name))
        tscat_model.do(SetAttributeAction(None, self._selected_entities(), self.old_name, values))

        self._select(self._selected_entities())


class DeleteAttribute(_EntityBased):
    def __init__(self, state: AppState, name: str,
                 parent=None):
        super().__init__(state, parent)

        self.setText(f'Delete attribute {name} from {self._select_state.type}')

        self.name = name
        self.values = [entity.__dict__[name] for entity in self._mapped_selected_entities()]

    def _redo(self) -> None:
        from .tscat_driver.model import tscat_model
        tscat_model.do(DeleteAttributeAction(None, self._selected_entities(), self.name))
        self._select(self._selected_entities())

    def _undo(self) -> None:
        from .tscat_driver.model import tscat_model
        tscat_model.do(SetAttributeAction(None, self._selected_entities(), self.name, self.values))
        self._select(self._selected_entities())


class SetAttributeValue(_EntityBased):
    def __init__(self, state: AppState, name: str, value,
                 parent=None):
        super().__init__(state, parent)

        self.setText(f'Change {name} to {value} in {self._select_state.type}')

        self.name = name
        self.previous_values = [entity.__dict__[name] for entity in self._mapped_selected_entities()]
        self.new_value = value

    def _redo(self) -> None:
        from .tscat_driver.model import tscat_model
        tscat_model.do(SetAttributeAction(None, self._selected_entities(), self.name,
                                          [self.new_value] * len(self._selected_entities())))

        self._select(self._selected_entities())

    def _undo(self) -> None:
        from .tscat_driver.model import tscat_model
        tscat_model.do(SetAttributeAction(None, self._selected_entities(), self.name, self.previous_values))

        self._select(self._selected_entities())


class NewCatalogue(_EntityBased):
    def __init__(self, state: AppState, parent=None) -> None:
        super().__init__(state, parent)

        self.setText('Create new Catalogue')

        self.uuid: Optional[str] = None

    def _redo(self) -> None:
        def creation_callback(action: CreateEntityAction) -> None:
            print("New Catalogue created", action.entity)
            assert action.entity is not None
            self.uuid = action.entity.uuid
            assert self.uuid is not None
            self._select([self.uuid], tscat._Catalogue)

        from .tscat_driver.model import tscat_model
        tscat_model.do(CreateEntityAction(creation_callback, tscat._Catalogue,
                                          {
                                              'name': "New Catalogue",
                                              'author': os.getlogin(),
                                              'uuid': self.uuid,
                                              PathAttributeName: self.state.current_catalogue_path()
                                          }))

    def _undo(self) -> None:
        def remove_callback(_: RemoveEntitiesAction) -> None:
            self._select(self._selected_entities())

        from .tscat_driver.model import tscat_model
        assert self.uuid is not None
        tscat_model.do(RemoveEntitiesAction(remove_callback, [self.uuid], True))


class NewEvent(_EntityBased):
    def __init__(self, state: AppState, parent=None):
        super().__init__(state, parent)

        self.setText('Create new Event')

        self.uuid: Optional[str] = None

    def _redo(self):
        from .tscat_driver.model import tscat_model

        def add_to_catalogue_callback(action: AddEventsToCatalogueAction) -> None:
            assert self.uuid is not None
            self._select([self.uuid], tscat._Event)

        def creation_callback(action: CreateEntityAction) -> None:
            self.uuid = action.entity.uuid
            assert self.uuid is not None  # satisfy mypy
            tscat_model.do(AddEventsToCatalogueAction(add_to_catalogue_callback,
                                                      [self.uuid], self._select_state.selected_catalogues[0]))

        tscat_model.do(CreateEntityAction(creation_callback, tscat._Event,
                                          {
                                              'start': dt.datetime.now(),
                                              'stop': dt.datetime.now(),
                                              'author': os.getlogin(),
                                              'uuid': self.uuid
                                          }))

    def _undo(self):
        def remove_callback(_: RemoveEntitiesAction) -> None:
            self._select(self._selected_entities())

        from .tscat_driver.model import tscat_model
        tscat_model.do(RemoveEntitiesAction(remove_callback, [self.uuid], True))


class MoveRestoreTrashedEntity(_EntityBased):
    def __init__(self, state: AppState, parent=None):
        super().__init__(state, parent)

    def remove(self):
        from .tscat_driver.model import tscat_model
        tscat_model.do(MoveToTrashAction(None, self._selected_entities()))

        self._select(self._selected_entities())

    def restore(self):
        from .tscat_driver.model import tscat_model
        tscat_model.do(RestoreFromTrashAction(None, self._selected_entities()))
        self._select(self._selected_entities())


class MoveEntityToTrash(MoveRestoreTrashedEntity):
    def __init__(self, state: AppState, parent=None):
        super().__init__(state, parent)

        self.setText(f'Move {self._select_state.type} to Trash')

    def _redo(self):
        self.remove()

    def _undo(self):
        self.restore()


class RestoreEntityFromTrash(MoveRestoreTrashedEntity):
    def __init__(self, state: AppState, parent=None):
        super().__init__(state, parent)

        self.setText(f'Restore {self._select_state.type} from Trash')

    def _redo(self):
        self.restore()

    def _undo(self):
        self.remove()


class DeletePermanently(_EntityBased):
    def __init__(self, state: AppState, parent=None):
        super().__init__(state, parent)

        # keep the action and its data to use it in case of undo
        self._deleted_entities: List[_DeletedEntity] = []

        self.setText(f'Delete {self._select_state.type} permanently')

    def _redo(self):
        def action_callback(action: DeletePermanentlyAction) -> None:
            self._deleted_entities = action.deleted_entities
            self._select([])

        from .tscat_driver.model import tscat_model
        tscat_model.do(DeletePermanentlyAction(action_callback, self._selected_entities()))

    def _undo(self):
        def action_callback(action: RestorePermanentlyDeletedAction) -> None:
            restored_uuids = [e.data['uuid'] for e in action.deleted_entities]
            self._select(restored_uuids)

        from .tscat_driver.model import tscat_model
        tscat_model.do(RestorePermanentlyDeletedAction(action_callback, self._deleted_entities))


class Import(_EntityBased):
    def __init__(self, state: AppState, filename: str, canonicalized_import_dict: dict, parent=None):
        super().__init__(state, parent)

        self.setText(f"Importing catalogues and events from {filename}.")

        self.import_dict = canonicalized_import_dict

    def _redo(self):
        from .tscat_driver.model import tscat_model
        tscat_model.do(ImportCanonicalizedDictAction(None, deepcopy(self.import_dict)))

    def _undo(self):
        from .tscat_driver.model import tscat_model

        event_uuids = [uuid for uuid in self.import_dict.events.keys()]
        tscat_model.do(RemoveEntitiesAction(None, event_uuids, True))

        catalogue_uuids = [cat["uuid"] for cat in self.import_dict.catalogues]
        tscat_model.do(RemoveEntitiesAction(None, catalogue_uuids, True))


class AddEventsToCatalogue(_EntityBased):
    def __init__(self, state: AppState,
                 catalogue_uuid: str, event_uuids: List[str],
                 parent=None):
        super().__init__(state, parent)

        self.setText(f'Add events to catalogue')
        self.catalogue_uuid = catalogue_uuid

        # when adding events to a catalogue events might already be present or indirectly selected by the predicate
        # so in the first redo-call we add all events, the driver will check whether events are already present and
        # return which ones have really been added
        self.event_uuids = event_uuids

    def _redo(self):
        def update_added_events(action: AddEventsToCatalogueAction):
            self.event_uuids = action.uuids
            self._select_state.selected_catalogues = [self.catalogue_uuid]
            self._select(self.event_uuids, type=tscat._Event)

        from .tscat_driver.model import tscat_model
        tscat_model.do(AddEventsToCatalogueAction(update_added_events, self.event_uuids, self.catalogue_uuid))

    def _undo(self):
        from .tscat_driver.model import tscat_model
        tscat_model.do(RemoveEventsFromCatalogueAction(None, self.event_uuids, self.catalogue_uuid))

        self._select([self.catalogue_uuid], type=tscat._Catalogue)


class CreateOrSetCataloguePath(_EntityBased):
    def __init__(self, state: AppState,
                 catalogues_paths: Dict[str, Tuple[List[str]]], parent=None):
        super().__init__(state, parent)

        self._select_state.selected_catalogues = list(catalogues_paths.keys())
        self.setText('Move catalogues to new paths')
        self.paths = catalogues_paths

    def _redo(self) -> None:
        def action_callback(_: SetAttributeAction) -> None:
            self._select(self._selected_entities())

        from .tscat_driver.model import tscat_model
        uuids, paths = [], []
        for u, p in self.paths.items():
            uuids.append(u)
            paths.append(p[1])
        tscat_model.do(SetAttributeAction(action_callback, uuids, PathAttributeName, paths))

    def _undo(self) -> None:
        def action_callback(_: SetAttributeAction) -> None:
            self._select(self._selected_entities())

        from .tscat_driver.model import tscat_model
        uuids, paths = [], []
        for u, p in self.paths.items():
            uuids.append(u)
            paths.append(p[0])
        tscat_model.do(SetAttributeAction(action_callback, uuids, PathAttributeName, paths))



import pickle
from typing import Any, Dict, List, Optional, Sequence, Set, Union, cast

from PySide6.QtCore import QAbstractItemModel, QMimeData, QModelIndex, QPersistentModelIndex, Qt, Signal
from PySide6.QtWidgets import QApplication, QStyle

from tscat import _Catalogue
from .actions import Action, CreateEntityAction, DeleteAttributeAction, DeletePermanentlyAction, GetCatalogueAction, \
    GetCataloguesAction, ImportCanonicalizedDictAction, MoveToTrashAction, RemoveEntitiesAction, \
    RestoreFromTrashAction, RestorePermanentlyDeletedAction, SetAttributeAction
from .catalog_model import CatalogModel
from .driver import tscat_driver
from .nodes import CatalogNode, FolderNode, NamedNode, Node, RootNode, TrashNode
from ..model_base.constants import EntityRole, PathAttributeName, UUIDDataRole
from .mime import is_catalogue_uuid_list, is_event_uuid_list, catalogue_uuid_list, event_uuid_list, MimeType


class TscatRootModel(QAbstractItemModel):
    events_dropped_on_catalogue = Signal(str, list)
    catalogues_dropped_on_folder = Signal(dict)

    def __init__(self) -> None:
        super().__init__()
        self._root = RootNode()
        self._trash = TrashNode()

        self._root.append_child(self._trash)

        self._catalogues: Dict[str, CatalogModel] = {}

        self.expanded_indexes: Set[QModelIndex] = set()

        tscat_driver.action_done_prioritised.connect(self._driver_action_done, Qt.QueuedConnection)

        tscat_driver.do(GetCataloguesAction(None, False))
        tscat_driver.do(GetCataloguesAction(None, True))

    def collapsed(self, index: QModelIndex) -> None:
        self.expanded_indexes.remove(index)
        self.dataChanged.emit(index, Qt.DecorationRole)  # type: ignore

    def expanded(self, index: QModelIndex) -> None:
        self.expanded_indexes.add(index)
        self.dataChanged.emit(index, Qt.DecorationRole)  # type: ignore

    def _trash_index(self) -> QModelIndex:
        return self.index(0, 0, QModelIndex())

    def node_from_path(self, path: List[str], create_index_for_new_folders: bool) -> Node:
        parent = self._root
        for folder in path:
            for child in parent.children:
                if isinstance(child, FolderNode) and child.name == folder:
                    parent = child
                    break
            else:
                if create_index_for_new_folders:
                    parent_index = self._index_from_node(parent)
                    self.beginInsertRows(parent_index, len(parent.children), len(parent.children))

                new_folder = FolderNode(folder)
                parent.append_child(new_folder)
                parent = new_folder

                if create_index_for_new_folders:
                    self.endInsertRows()

        return parent

    def _node_from_catalogue_path(self, c: _Catalogue, create_index_for_new_folders: bool = False) -> Node:
        parent = self._root
        if hasattr(c, PathAttributeName) and isinstance(getattr(c, PathAttributeName), list) and \
            all(isinstance(x, str) for x in getattr(c, PathAttributeName)):
            parent = self.node_from_path(getattr(c, PathAttributeName), create_index_for_new_folders)

        return parent

    def _node_from_uuid(self, uuid: str) -> Optional[CatalogNode]:
        stack = [self._root]

        while stack:
            node = stack.pop()
            if isinstance(node, (CatalogNode, FolderNode)) and node.uuid == uuid:
                return node

            # we do not want to search in Catalogue's children, they are events
            if not isinstance(node, CatalogNode):
                stack.extend(node.children)

        return None

    def _index_from_node(self, node: Node) -> QModelIndex:
        if node.parent is None:
            return QModelIndex()

        return self.index(node.row, 0, self._index_from_node(node.parent))

    def _get_node_from_uuid_and_remove_from_tree(self, uuid: str) -> Optional[CatalogNode]:
        node = self._node_from_uuid(uuid)
        if node is None:
            return None

        index = self._index_from_node(node)

        self.beginRemoveRows(index.parent(), index.row(), index.row())
        node.parent.remove_child(node)
        self.endRemoveRows()

        return node

    def _insert_catalogue_node_at_node_path_or_trash(self, node: CatalogNode) -> None:
        if node.node.is_removed():  # undelete to Trash
            parent_node = self._trash
        else:
            parent_node = self._node_from_catalogue_path(node.node, create_index_for_new_folders=True)
        parent_index = self._index_from_node(parent_node)

        self.beginInsertRows(parent_index, len(parent_node.children), len(parent_node.children))
        parent_node.append_child(node)
        self.endInsertRows()

    @staticmethod
    def _is_inside_trash(index: QModelIndex) -> bool:
        if not index.isValid():
            return False

        # check if the index is inside the TrashNode
        node = cast(Node, index.internalPointer())
        while node is not None:
            if isinstance(node, TrashNode):
                return True
            node = node.parent

        return False

    def index_from_uuid(self, uuid: str) -> QModelIndex:
        node = self._node_from_uuid(uuid)
        if node is None:
            return QModelIndex()
        return self._index_from_node(node)

    def _driver_action_done(self, action: Action) -> None:
        if isinstance(action, GetCataloguesAction):
            if action.removed_items:
                self.beginRemoveRows(self._trash_index(), 0, len(self._trash.children) - 1)
                self._trash.set_children([])
                self.endRemoveRows()

                self.beginInsertRows(self._trash_index(), 0, len(action.catalogues) - 1)
                self._trash.set_children(list(map(CatalogNode, action.catalogues)))
                self.endInsertRows()
            else:
                self.beginResetModel()
                self._root.set_children([self._trash])

                for c in action.catalogues:
                    parent_node = self._node_from_catalogue_path(c)
                    parent_node.append_child(CatalogNode(c))
                self.endResetModel()

        elif isinstance(action, GetCatalogueAction):
            # also search in Trash :D
            index = self.index_from_uuid(action.uuid)
            if index.isValid():
                self.dataChanged.emit(index, index)  # type: ignore

        elif isinstance(action, CreateEntityAction):
            if isinstance(action.entity, _Catalogue):
                node = CatalogNode(action.entity)
                self._insert_catalogue_node_at_node_path_or_trash(node)

        elif isinstance(action, (RemoveEntitiesAction, DeletePermanentlyAction)):
            for uuid in action.uuids:
                node = self._get_node_from_uuid_and_remove_from_tree(uuid)
                if node is None:
                    continue

                # in case a catalogueModel was created for this catalogue, remove it
                if uuid in self._catalogues:
                    self._catalogues[uuid].deleteLater()
                    del self._catalogues[uuid]

        elif isinstance(action, (MoveToTrashAction, RestoreFromTrashAction)):  # Move to Trash exists only for Catalogs
            for uuid, entity in zip(action.uuids, action.entities):
                node = self._get_node_from_uuid_and_remove_from_tree(uuid)
                if node is None:
                    continue

                node.node = entity  # update the node's entity to the new one with the removed flag set

                self._insert_catalogue_node_at_node_path_or_trash(node)

        elif isinstance(action, RestorePermanentlyDeletedAction):
            for e in action.deleted_entities:
                if isinstance(e.restored_entity, _Catalogue):
                    node = CatalogNode(e.restored_entity)
                    self._insert_catalogue_node_at_node_path_or_trash(node)

        elif isinstance(action, (SetAttributeAction, DeleteAttributeAction)):
            for c in filter(lambda x: isinstance(x, _Catalogue), action.entities):
                node = self._node_from_uuid(c.uuid)
                if node is not None:
                    node.node = c

                    if action.name == PathAttributeName:
                        node = self._get_node_from_uuid_and_remove_from_tree(c.uuid)
                        self._insert_catalogue_node_at_node_path_or_trash(node)

                    index = self._index_from_node(node)
                    self.dataChanged.emit(index, index)  # type: ignore

        elif isinstance(action, ImportCanonicalizedDictAction):
            for node in list(map(CatalogNode, action.catalogues)):
                self._insert_catalogue_node_at_node_path_or_trash(node)

    def catalog(self, uuid: str) -> CatalogModel:
        if uuid not in self._catalogues:
            # DFS search for the CatalogNode with the given uuid
            stack = [self._root]

            while stack:
                node = stack.pop()
                if isinstance(node, CatalogNode) and node.uuid == uuid:
                    catalogue_model = CatalogModel(node)
                    self._catalogues[uuid] = catalogue_model
                    tscat_driver.do(GetCatalogueAction(None, removed_items=False, uuid=uuid))
                    break

                if not isinstance(node, CatalogNode):  # we do not want to search in Catalogue's children
                    stack.extend(node.children)

        return self._catalogues[uuid]

    def current_path(self, index: QModelIndex) -> List[str]:
        if not index.isValid():
            return []

        # get the path starting with the parent if the index is a catalogue else starting with the index
        if index.data(EntityRole) is not None:
            index = index.parent()

        path = []

        while index.isValid():
            path.append(index.data())
            index = index.parent()
        return path[::-1]

    def index(self, row: int, column: int,
              parent: Union[QModelIndex, QPersistentModelIndex] = QModelIndex()) -> QModelIndex:
        if self.hasIndex(row, column, parent):
            if not parent.isValid():
                parent_item: Node = self._root
            else:
                parent_item: Node = parent.internalPointer()  # type: ignore

            child_item: Node = parent_item.children[row]
            if child_item is not None:
                return self.createIndex(row, column, child_item)
        return QModelIndex()

    def parent(self, index: Union[QModelIndex, QPersistentModelIndex]) -> QModelIndex:  # type: ignore
        if not index.isValid():
            return QModelIndex()
        assert isinstance(index.internalPointer(), Node)
        child_item = cast(Node, index.internalPointer())
        assert isinstance(child_item.parent, Node)
        parent_item: Node = child_item.parent
        if parent_item not in (None, self._root):
            return self.createIndex(parent_item.row, 0, parent_item)
        return QModelIndex()

    def rowCount(self, parent: Union[QModelIndex, QPersistentModelIndex] = QModelIndex()) -> int:  # type: ignore
        if parent.column() > 0:
            return 0
        parent_node: Node
        if not parent.isValid():
            parent_node = self._root
        else:
            parent_node = cast(Node, parent.internalPointer())

        if isinstance(parent_node, CatalogNode):
            return 0
        else:
            return len(parent_node.children)

    def columnCount(self, parent: Union[QModelIndex, QPersistentModelIndex] = None) -> int:  # type: ignore
        return 1

    def data(self, index: Union[QModelIndex, QPersistentModelIndex],
             role: Qt.ItemDataRole = Qt.DisplayRole) -> Any:  # type: ignore
        if index.isValid():
            item = cast(NamedNode, index.internalPointer())
            if role == Qt.ItemDataRole.DisplayRole:
                return item.name
            elif role == Qt.DecorationRole:
                if isinstance(item, CatalogNode):
                    return QApplication.style().standardIcon(QStyle.SP_FileIcon)
                elif isinstance(item, FolderNode):
                    if index in self.expanded_indexes:
                        return QApplication.style().standardIcon(QStyle.SP_DirOpenIcon)
                    else:
                        return QApplication.style().standardIcon(QStyle.SP_DirClosedIcon)
                elif isinstance(item, TrashNode):
                    return QApplication.style().standardIcon(QStyle.SP_TrashIcon)

            elif role == UUIDDataRole:
                if isinstance(item, (CatalogNode, FolderNode)):
                    return item.uuid
            elif role == EntityRole:
                if isinstance(item, CatalogNode):
                    return item.node
        return None

    def flags(self, index: Union[QModelIndex, QPersistentModelIndex]) -> Qt.ItemFlag:
        if index.isValid():
            item = cast(Node, index.internalPointer())
            return item.flags()
        return Qt.ItemFlag.NoItemFlags

    def headerData(self, section: int, orientation: Qt.Orientation, role: int = Qt.DisplayRole) -> Any:  # type: ignore
        if role == Qt.DisplayRole:  # type: ignore
            return "Catalogues"
        return None

    def supportedDragActions(self) -> Qt.DropAction:
        return Qt.DropAction.CopyAction

    def canDropMimeData(self, data: QMimeData, action: Qt.DropAction,
                        row: int, column: int,
                        parent: Union[QModelIndex, QPersistentModelIndex]) -> bool:

        if self._is_inside_trash(parent):
            return False
        # on Catalogues, we can drop events or catalogues (convenient for moving catalogues next to each other)
        elif isinstance(parent.data(EntityRole), _Catalogue):
            return  is_event_uuid_list(data) or is_catalogue_uuid_list(data)
        # on everything else except Trash we can drop catalogues - because it is a FolderNode or Trash
        else:
            return (parent != self._trash_index()) and is_catalogue_uuid_list(data)

    def dropMimeData(self, data: QMimeData, action: Qt.DropAction,
                     row: int, column: int,
                     parent: Union[QModelIndex, QPersistentModelIndex]) -> bool:
        if not self.canDropMimeData(data, action, row, column, parent):
            return False

        if action == Qt.DropAction.IgnoreAction:
            return True

        if is_event_uuid_list(data):
            self.events_dropped_on_catalogue.emit(parent.data(UUIDDataRole),
                                                  event_uuid_list(data))
        elif is_catalogue_uuid_list(data):  # per canDropMimeData, parent is a FolderNode
            uuids = catalogue_uuid_list(data)
            parent_path = self.current_path(parent)

            # dict keeping the old and new paths of the catalogues in a tuple, keyed by uuid
            catalogues_paths = {}
            for uuid in uuids:
                index = self.index_from_uuid(uuid)

                assert index.isValid()

                all_parents = [parent]
                i = parent
                while i.parent().isValid():
                    all_parents.append(i.parent())
                    i = i.parent()

                # do not allow dropping of path to itself or any of its children
                if index in all_parents:
                    continue

                if index.parent() == parent:  # if the parent is the same, no need to change the path
                    continue

                node = index.internalPointer()
                if isinstance(node, CatalogNode):  # for catalogues the new path is the parent's path
                    catalogues_paths[uuid] = (self.current_path(self._index_from_node(node)), parent_path)
                elif isinstance(node, FolderNode):
                    # for folders the new path for all catalogues below, is the
                    # parent's path + relative path starting at the dragged folder
                    catalogues = self._catalogue_nodes(node)
                    if not catalogues:  # folder with no catalogues in it
                        pass
                    else:
                        for c in catalogues:
                            folder_path = node.full_path()
                            catalogue_path = self.current_path(self._index_from_node(c))
                            catalogues_paths[c.uuid] = (catalogue_path,
                                                        parent_path + catalogue_path[len(folder_path) - 1:])
            # print(json.dumps(catalogues_new_path, indent=4))
            if catalogues_paths:
                self.catalogues_dropped_on_folder.emit(catalogues_paths)

        return True

    def mimeTypes(self) -> List[str]:
        return super().mimeTypes() + ['text/uri-list']

    def mimeData(self, indexes: Sequence[QModelIndex]) -> QMimeData:
        mime_data = super().mimeData(indexes)
        mime_data.setData(MimeType.CATALOGUE_UUID_LIST.value, pickle.dumps([self.data(index, UUIDDataRole)
                                                                                   for index in indexes]))
        return mime_data

    @staticmethod
    def _catalogue_nodes(parent_node: Node) -> List[CatalogNode]:
        stack = list(parent_node.children[:])

        catalogues: List[CatalogNode] = []

        while stack:
            node = stack.pop()
            if isinstance(node, CatalogNode):
                catalogues.append(node)

            if not isinstance(node, (CatalogNode, TrashNode)):  # we do not want to search in Catalogue's children
                stack.extend(node.children)

        return catalogues

    def catalogue_nodes(self, in_trash: bool) -> Sequence[CatalogNode]:
        if in_trash:
            return self._catalogue_nodes(self._trash)
        else:
            return self._catalogue_nodes(self._root)
